from __future__ import absolute_import, division, print_function

from collections import OrderedDict

import numpy as np
import scipy as sp
from matplotlib import gridspec
from matplotlib import pyplot as plt
from sklearn.cluster import KMeans
from sklearn.metrics import (adjusted_mutual_info_score, adjusted_rand_score,
                             normalized_mutual_info_score, silhouette_score)
from sklearn.mixture import GaussianMixture
from sklearn.multiclass import OneVsRestClassifier
from sklearn.svm import SVC
from sklearn.utils.linear_assignment_ import linear_assignment

from odin import backend as K
from odin.utils import catch_warnings_ignore, ctext, one_hot
from odin.visual import (plot_colorbar, plot_figure, plot_scatter_heatmap,
                         to_axis)
from sisua.data.const import (MARKER_GENES, PROTEIN_PAIR_COMPARISON,
                              UNIVERSAL_RANDOM_SEED)
from sisua.data.utils import standardize_protein_name
from sisua.label_threshold import ProbabilisticEmbedding
from sisua.utils.visualization import (downsample_data, fast_pca, fast_scatter,
                                       fast_tsne, plot_evaluate_classifier)


# ===========================================================================
# Metrics
# ===========================================================================
def multi_label_adj_Rindex(label_bin_true, label_pred):
  assert label_bin_true.ndim == 2
  assert label_bin_true.shape[1] == len(np.unique(label_pred))
  n_classes = label_bin_true.shape[1]
  with catch_warnings_ignore(Warning):
    scores = []
    for y in label_bin_true.T:
      y = y.astype('int32')
      s = max(
          adjusted_rand_score(labels_true=y,
                              labels_pred=(label_pred == i).astype('int32'))
          for i in range(n_classes))
      scores.append(s)
  return scores


def unsupervised_clustering_accuracy(y, y_pred):
  """
  Unsupervised Clustering Accuracy
  Author: scVI
  https://github.com/YosefLab/scVI/blob/a585f7d096f04ab0d50cadfdf8c2c9f78d907c19/scvi/inference/posterior.py#L637
  """
  assert len(y_pred) == len(y)
  u = np.unique(np.concatenate((y, y_pred)))
  n_clusters = len(u)
  mapping = dict(zip(u, range(n_clusters)))
  reward_matrix = np.zeros((n_clusters, n_clusters), dtype=np.int64)
  for y_pred_, y_ in zip(y_pred, y):
    if y_ in mapping:
      reward_matrix[mapping[y_pred_], mapping[y_]] += 1
  cost_matrix = reward_matrix.max() - reward_matrix
  ind = linear_assignment(cost_matrix)
  return sum([reward_matrix[i, j] for i, j in ind]) * 1.0 / y_pred.size, ind


def clustering_scores(latent, labels, n_labels, prediction_algorithm='both'):
  """ Clustering Scores:

   * silhouette_score (higher is better, best is 1, worst is -1)
   * adjusted_rand_score (higher is better)
   * normalized_mutual_info_score (higher is better)
   * unsupervised_clustering_accuracy (higher is better)

  note: remember the order of returned value

  Parameters
  ----------
  labels : categorical labels (i.e. single classes or one-hot encoded)
  prediction_algorithm : {'knn', 'gmm', 'both'}
  """
  # simple normalization to 0-1, then pick the argmax
  if labels.ndim == 2:
    min_val = np.min(labels, axis=0, keepdims=True)
    max_val = np.max(labels, axis=0, keepdims=True)
    labels = (labels - min_val) / (max_val - min_val)
    labels = np.argmax(labels, axis=-1)

  if prediction_algorithm == 'knn':
    km = KMeans(n_labels, n_init=200, random_state=5218)
    labels_pred = km.fit_predict(latent)
  elif prediction_algorithm == 'gmm':
    gmm = GaussianMixture(n_labels, random_state=5218)
    gmm.fit(latent)
    labels_pred = gmm.predict(latent)
  elif prediction_algorithm == 'both':
    score1 = clustering_scores(latent,
                               labels,
                               n_labels=n_labels,
                               prediction_algorithm='knn')
    score2 = clustering_scores(latent,
                               labels,
                               n_labels=n_labels,
                               prediction_algorithm='gmm')
    return {k: (v + score2[k]) / 2 for k, v in score1.items()}
  else:
    raise ValueError("Not support for prediction_algorithm: '%s'" %
                     prediction_algorithm)
  #
  with catch_warnings_ignore(FutureWarning):
    asw_score = silhouette_score(latent, labels)
    ari_score = adjusted_rand_score(labels, labels_pred)
    nmi_score = normalized_mutual_info_score(labels, labels_pred)
    uca_score = unsupervised_clustering_accuracy(labels, labels_pred)[0]
  return dict(ASW=asw_score, ARI=ari_score, NMI=nmi_score, UCA=uca_score)


# ===========================================================================
# Visualization
# ===========================================================================
def streamline_classifier(Z_train,
                          y_train,
                          Z_test,
                          y_test,
                          labels_name,
                          mode='ovr',
                          title='',
                          plot_train_results=False,
                          show_plot=True,
                          return_figure=False):
  """ Return a dictionary of scores
  {
      F1micro=f1_micro * 100,
      F1macro=f1_macro * 100,
      F1weight=f1_weight * 100,
      F1_[classname]=...
  }
  """
  mode = mode.strip().lower()
  assert mode in ('ovr', 'ovo'), \
  "Only support ovr - one vs rest, ovo - one vs one; mode for streamline classifier"

  labels_name = [standardize_protein_name(i) for i in labels_name]

  results_train = {}
  results_test = {}
  labels_name = np.array(labels_name)

  with catch_warnings_ignore(FutureWarning):
    with catch_warnings_ignore(RuntimeWarning):
      n_classes = len(labels_name)
      # ====== preprocessing ====== #
      if y_train.ndim == 1 or y_train.shape[1] == 1:
        y_train = one_hot(y_train.ravel(), nb_classes=n_classes)
      if y_test.ndim == 1 or y_test.shape[1] == 1:
        y_test = one_hot(y_test.ravel(), nb_classes=n_classes)
      is_binary_classes = sorted(np.unique(
          y_train.astype('float32'))) == [0., 1.]
      # ====== not binary classes ====== #
      if not is_binary_classes:
        gmm = ProbabilisticEmbedding()
        gmm.fit(np.concatenate((y_train, y_test), axis=0))
        y_train = gmm.predict(y_train)
        y_test = gmm.predict(y_test)
      # kernel : 'linear', 'poly', 'rbf', 'sigmoid', 'precomputed'
      if mode == 'ovr':
        classifier = OneVsRestClassifier(SVC(
            kernel='linear', random_state=UNIVERSAL_RANDOM_SEED),
                                         n_jobs=n_classes)
        classifier.fit(X=Z_train, y=y_train)
      else:
        raise NotImplementedError
        classifier = SVC(kernel='linear',
                         decision_function_shape='ovo',
                         random_state=UNIVERSAL_RANDOM_SEED)
        classifier.fit(X=Z_train, y=y_train)
      # ====== return ====== #
      from sklearn.exceptions import UndefinedMetricWarning
      with catch_warnings_ignore(UndefinedMetricWarning):
        results_train = plot_evaluate_classifier(
            y_pred=classifier.predict(Z_train),
            y_true=y_train,
            labels=labels_name,
            title='[train]' + title,
            show_plot=show_plot and plot_train_results,
            return_figure=True)
        results_test = plot_evaluate_classifier(
            y_pred=classifier.predict(Z_test),
            y_true=y_test,
            labels=labels_name,
            title='[test]' + title,
            show_plot=show_plot,
            return_figure=True)

      if show_plot:
        if plot_train_results:
          results_train, fig_train = results_train[0], results_train[1]
        else:
          fig_train = None
        results_test, fig_test = results_test[0], results_test[1]
      results_train = OrderedDict(
          sorted(results_train.items(), key=lambda x: x[0]))
      results_test = OrderedDict(
          sorted(results_test.items(), key=lambda x: x[0]))
      results = (results_train, results_test)

      if show_plot and return_figure:
        return results, (fig_train, fig_test)
      return results


def plot_distance_heatmap(X,
                          labels,
                          labels_name=None,
                          lognorm=True,
                          colormap='hot',
                          ax=None,
                          legend_enable=True,
                          legend_loc='upper center',
                          legend_ncol=3,
                          legend_colspace=0.2,
                          fontsize=10,
                          show_colorbar=True,
                          title=None):
  """
  Parameters
  ----------
  X : (n_samples, n_features)
    coordination for scatter points

  labels : (n_samples, n_classes) or (n_samples, 1) or (n_samples,)
    list of classes index, in case of binary classification,
    the list can be float value represent confidence value for
    positive class.

  labels_name : (n_classes,)
    list of classes' name, this will be used to determine
    number of classes

  # visualize_distance(latent_scVI, labels, "scVI")
  """
  from matplotlib.lines import Line2D
  X = K.length_norm(X, axis=-1, epsilon=np.finfo(X.dtype).eps)

  ax = to_axis(ax)
  n_samples, n_dim = X.shape

  # processing labels
  labels = np.array(labels)
  if labels.ndim == 2:
    if labels.shape[1] == 1:
      labels = labels.ravel()
    else:
      labels = np.argmax(labels, axis=-1)
  elif labels.ndim > 2:
    raise ValueError("Only support 1-D or 2-D labels")

  labels_int = labels.astype('int32')
  # float values label (normalize -1 to 1) or binary classification
  if not np.all(labels_int == labels) or \
  (labels_name is not None and len(labels_name) == 2) or \
  (len(np.unique(labels)) == 2):
    min_val = np.min(labels)
    max_val = np.max(labels)
    labels = 2 * (labels - min_val) / (max_val - min_val) - 1
    label_colormap = 'bwr'
  # integer values label and multiple classes classification
  else:
    labels = labels_int
    label_colormap = 'Dark2'

  # ====== sorting label and X ====== #
  order_X = np.vstack(
      [x for _, x in sorted(zip(labels, X), key=lambda pair: pair[0])])
  order_label = np.vstack(
      [y for y, x in sorted(zip(labels, X), key=lambda pair: pair[0])])
  distance = sp.spatial.distance_matrix(order_X, order_X)
  if bool(lognorm):
    distance = np.log1p(distance)
  min_non_zero = np.min(distance[np.nonzero(distance)])
  distance = np.clip(distance, a_min=min_non_zero, a_max=np.max(distance))

  # ====== convert data to image ====== #
  cm = plt.get_cmap(colormap)
  distance_img = cm(distance)
  # diagonal black line (i.e. zero distance)
  for i in range(n_samples):
    distance_img[i, i] = (0, 0, 0, 1)

  cm = plt.get_cmap(label_colormap)
  width = max(int(0.032 * n_samples), 8)
  horz_bar = np.repeat(cm(order_label.T), repeats=width, axis=0)
  vert_bar = np.repeat(cm(order_label), repeats=width, axis=1)

  final_img = np.zeros(shape=(n_samples + width, n_samples + width,
                              distance_img.shape[2]),
                       dtype=distance_img.dtype)
  final_img[width:, width:] = distance_img
  final_img[:width, width:] = horz_bar
  final_img[width:, :width] = vert_bar
  assert np.sum(final_img[:width, :width]) == 0, \
  "Something wrong with my spacial coordination when writing this code!"
  # ====== plotting ====== #
  ax.imshow(final_img)
  ax.axis('off')
  # ====== legend ====== #
  if labels_name is not None and bool(legend_enable):
    cm = plt.get_cmap(label_colormap)
    labels_name = np.asarray(labels_name)
    if len(labels_name) == 2:  # binary (easy peasy)
      all_colors = np.array((cm(np.min(labels)), cm(np.max(labels))))
    else:  # multiple classes
      all_colors = cm(list(range(len(labels_name))))
    legend_elements = [
        Line2D([0], [0],
               marker='o',
               color=color,
               label=name,
               linewidth=0,
               linestyle=None,
               lw=0,
               markerfacecolor=color,
               markersize=fontsize // 2)
        for color, name in zip(all_colors, labels_name)
    ]
    ax.legend(handles=legend_elements,
              markerscale=1.,
              scatterpoints=1,
              scatteryoffsets=[0.375, 0.5, 0.3125],
              loc=legend_loc,
              bbox_to_anchor=(0.5, -0.01),
              ncol=int(legend_ncol),
              columnspacing=float(legend_colspace),
              labelspacing=0.,
              fontsize=fontsize - 1,
              handletextpad=0.1)
  # ====== final configurations ====== #
  if title is not None:
    ax.set_title(str(title), fontsize=fontsize)
  if show_colorbar:
    plot_colorbar(colormap,
                  vmin=np.min(distance),
                  vmax=np.max(distance),
                  ax=ax,
                  orientation='vertical')
  return ax


def plot_latents_multiclasses(Z,
                              y,
                              labels_name,
                              title=None,
                              elev=None,
                              azim=None,
                              use_PCA=False,
                              show_colorbar=False):
  """ Label `y` is multi-classes
  i.e. each samples could belong to multiple classes at once

  Return
  ------
  fig : matplotlib.Figure or None
      if no pair found, return None, otherwise, the
      figure used to plot all protein pairs

  """
  from sisua.data.utils import standardize_protein_name
  labels_name = [standardize_protein_name(i) for i in labels_name]

  if title is None:
    title = ''
  title = '[%s]%s' % ("PCA" if use_PCA else "t-SNE", title)
  # ====== Downsample if the data is huge ====== #
  Z, y = downsample_data(Z, y)
  # ====== checking inputs ====== #
  assert Z.ndim == 2, Z.shape
  assert Z.shape[0] == y.shape[0]
  # ====== preprocessing ====== #
  if Z.shape[1] > 3:
    if not use_PCA:
      Z = fast_tsne(Z,
                    n_components=2,
                    perplexity=30.0,
                    learning_rate=200,
                    n_iter=1000,
                    random_state=87654321,
                    n_jobs=8)
    else:
      Z = fast_pca(Z, n_components=2, random_state=87654321)

  # ====== select proteins ====== #
  def logit(p):
    eps = np.finfo('float32').eps
    p = np.copy(p)
    p[p == 0] = eps
    p[p == 1] = 1 - eps
    return np.log(p / (1 - p))

  # normalize to 0, 1
  y_min = np.min(y, axis=0, keepdims=True)
  y_max = np.max(y, axis=0, keepdims=True)
  y = (y - y_min) / (y_max - y_min)

  # select most 2 different protein
  labels_index = {name: i for i, name in enumerate(labels_name)}
  pairs = []
  for i, j in PROTEIN_PAIR_COMPARISON:
    if i in labels_index and j in labels_index:
      pairs.append((i, j))
  n_pairs = len(pairs)

  if n_pairs == 0:
    return None

  # we could handle 5 pairs in 1 row, no problem
  ncol = min(5, n_pairs)
  nrow = int(np.ceil(n_pairs / ncol))
  fig = plt.figure(figsize=(ncol * 4, nrow * 4))

  for idx, labels_name in enumerate(pairs):
    ax = plt.subplot(nrow, ncol, idx + 1)
    # polarize y level
    val = np.hstack((y[:, labels_index[labels_name[0]]][:, np.newaxis],
                     y[:, labels_index[labels_name[1]]][:, np.newaxis]))
    # red mean closer to 1, i.e. protein labels_name[1]
    # blue mean closer to -1, i.e. protein labels_name[0]
    val = logit(val[:, 1]) - logit(val[:, 0])
    # normalize again to [-1, 1]
    val = 2 * (val - np.min(val)) / (np.max(val) - np.min(val)) - 1
    # ====== let plotting ====== #
    plot_scatter_heatmap(
        x=Z[:, 0],
        y=Z[:, 1],
        val=val,
        legend_enable=False,
        colormap='bwr',
        size=8,
        alpha=1.,
        fontsize=8,
        grid=False,
        ax=ax,
        colorbar=True,
        colorbar_horizontal=True,
        colorbar_ticks=[labels_name[0], 'Others', labels_name[1]],
        title='%s' % ('/'.join(labels_name)))
  plt.suptitle(title)
  return fig


def plot_latents_binary(Z,
                        y,
                        labels_name,
                        title=None,
                        elev=None,
                        azim=None,
                        use_PCA=False,
                        ax=None,
                        show_legend=True,
                        size=12,
                        fontsize=12,
                        show_scores=True,
                        enable_argmax=True,
                        enable_separated=False):
  from matplotlib import pyplot as plt
  if title is None:
    title = ''
  title = '[%s]%s' % ("PCA" if use_PCA else "t-SNE", title)
  ax = to_axis(ax)
  # ====== Downsample if the data is huge ====== #
  Z, y = downsample_data(Z, y)
  # ====== checking inputs ====== #
  assert Z.ndim == 2, Z.shape
  assert Z.shape[0] == y.shape[0]
  num_classes = len(labels_name)
  # ====== preprocessing ====== #
  if Z.shape[1] > 3:
    if not use_PCA:
      Z = fast_tsne(Z,
                    n_components=2,
                    perplexity=30.0,
                    learning_rate=200,
                    n_iter=1000,
                    random_state=87654321,
                    n_jobs=8)
    else:
      Z = fast_pca(Z, n_components=2, random_state=87654321)
  # ====== clustering metrics ====== #
  if show_scores:
    scores = clustering_scores(latent=Z,
                               labels=np.argmax(y, axis=-1) if y.ndim == 2 else y,
                               n_labels=num_classes)
    title += '\n'
    for k, v in sorted(scores.items(), key=lambda x: x[0]):
      title += '%s:%.2f ' % (k, v)
  # ====== plotting ====== #
  if enable_argmax:
    y_argmax = np.argmax(y, axis=-1) if y.ndim == 2 else y
    fast_scatter(x=Z,
                 y=y_argmax,
                 labels=labels_name,
                 ax=ax,
                 size=size,
                 title=title,
                 fontsize=fontsize,
                 enable_legend=bool(show_legend))
    ax.grid(False)
  # ====== plot each protein ====== #
  if enable_separated:
    colormap = 'Reds'  # bwr
    ncol = 5 if num_classes <= 20 else 9
    nrow = int(np.ceil(num_classes / ncol))
    fig = plot_figure(nrow=4 * nrow, ncol=20)
    for i, lab in enumerate(labels_name):
      val = K.log_norm(y[:, i], axis=0)
      plot_scatter_heatmap(x=Z[:, 0],
                           y=Z[:, 1],
                           val=val / np.sum(val),
                           ax=(nrow, ncol, i + 1),
                           colormap=colormap,
                           size=size,
                           alpha=0.8,
                           fontsize=8,
                           grid=False,
                           title=lab)

    plt.grid(False)
    # big title
    plt.suptitle(title, fontsize=fontsize)
    # show the colorbar
    import matplotlib as mpl
    cbar_ax = fig.add_axes([0.92, 0.15, 0.02, 0.7])
    cmap = mpl.cm.get_cmap(name=colormap)
    norm = mpl.colors.Normalize(vmin=0., vmax=1.)
    cb1 = mpl.colorbar.ColorbarBase(cbar_ax,
                                    cmap=cmap,
                                    norm=norm,
                                    orientation='vertical')
    cb1.set_label('Protein markers level')



from __future__ import absolute_import, division, print_function

from typing import Iterable, List, Text, Union

import tensorflow as tf
from tensorflow.python.keras.layers import Layer

from odin.bay import Statistic
from odin.bay.distribution_alias import parse_distribution
from odin.networks import DenseDistribution, Parallel
from odin.utils import as_tuple
from sisua.models.autoencoder import DeepCountAutoencoder
from sisua.models.base import SingleCellModel
from sisua.models.scvi_models import SCVI
from sisua.models.variational_autoencoder import VariationalAutoEncoder



class MultitaskAutoEncoder(DeepCountAutoencoder):
  """ Multitask Auto Encoder
  """

  def __init__(self,
               units,
               dispersion='full',
               xdist=['zinb', 'nb'],
               xdrop=0.3,
               edrop=0,
               zdrop=0,
               ddrop=0,
               hdim=128,
               zdim=32,
               biased_latent=False,
               nlayers=2,
               batchnorm=True,
               linear_decoder=False,
               **kwargs):
    kw = dict(locals())
    del kw['self']
    del kw['__class__']
    del kw['kwargs']
    kw.update(kwargs)
    super(MultitaskAutoEncoder, self).__init__(**kw)


class MultitaskVAE(VariationalAutoEncoder):
  """ Variational autoencoder with multitask learning extension
  """

  def __init__(self,
               units,
               dispersion='full',
               xdist=['zinb', 'nb'],
               zdist='normal',
               xdrop=0.3,
               edrop=0,
               zdrop=0,
               ddrop=0,
               hdim=128,
               zdim=32,
               nlayers=2,
               batchnorm=True,
               linear_decoder=False,
               **kwargs):
    kw = dict(locals())
    del kw['self']
    del kw['__class__']
    del kw['kwargs']
    kw.update(kwargs)
    super(MultitaskVAE, self).__init__(**kw)


class MultitaskVI(SCVI):
  """ Semi-supervised implementation of scVI
  """

  def __init__(self,
               units,
               dispersion='full',
               xdist=['zinbd', 'nb'],
               zdist='normal',
               ldist='normal',
               xdrop=0.3,
               edrop=0,
               zdrop=0,
               ddrop=0,
               hdim=128,
               zdim=32,
               nlayers=2,
               clip_library=12,
               batchnorm=True,
               linear_decoder=False,
               **kwargs):
    kw = dict(locals())
    del kw['self']
    del kw['__class__']
    del kw['kwargs']
    kw.update(kwargs)
    super(MultitaskVI, self).__init__(**kw)


# class MultiLatentVAE(VariationalAutoEncoder):
#   pass



from __future__ import absolute_import, division, print_function

import os

import numpy as np
from six import string_types

from odin.utils import ctext


# ===========================================================================
# For saving data
# ===========================================================================
def save_data(outpath, header, row, data):
  """ Simple shortcut

  If `feather-format` is available, save data to `.feather` file
  otherwise `.csv`
  """
  try:
    import feather
    return save_data_to_R(outpath, header, row, data)
  except ImportError as e:
    return save_data_to_csv(outpath, header, row, data)


def save_data_to_csv(outpath, header, row, data):
  if data is None:
    return
  if '.csv' not in outpath:
    outpath += '.csv'

  assert len(row) == len(data), "Data length mismatch!"
  with open(outpath, 'w') as f:
    # preprocessing the head
    if isinstance(header, (tuple, list, np.ndarray, set)):
      header = ",".join([
          str(val) if isinstance(val, (string_types, np.string_)) else "D%d" % i
          for i, val in enumerate(header)
      ])
    f.write('Cell,' + header + '\n')
    # write the data
    for i, j in zip(row, data):
      f.write(i + ',')
      f.write(','.join(['%g' % x for x in j]))
      f.write('\n')


def save_data_to_R(outpath, header, row, data):
  if data is None:
    return
  if '.feather' not in outpath:
    outpath += '.feather'

  import pandas as pd
  try:
    import feather
  except ImportError as e:
    raise RuntimeError(
        "Cannot export to R, require python package 'feather-format'")

  row = np.array(row)
  if isinstance(header, string_types):
    header = header.split(',')
  header = np.array(header)
  df = pd.DataFrame(data=data, index=row, columns=header, dtype=data.dtype)

  feather.write_dataframe(df, outpath)



from __future__ import absolute_import, division, print_function

import gzip
import os
import pickle
import shutil
import struct

import numpy as np

from odin import fuel as F
from odin.utils import ctext, get_file, one_hot, select_path
from sisua.data.path import DOWNLOAD_DIR, PREPROCESSED_BASE_DIR

_URLs = {
    "values": {
        "training":
            "http://yann.lecun.com/exdb/mnist/train-images-idx3-ubyte.gz",
        "test":
            "http://yann.lecun.com/exdb/mnist/t10k-images-idx3-ubyte.gz"
    },
    "labels": {
        "training":
            "http://yann.lecun.com/exdb/mnist/train-labels-idx1-ubyte.gz",
        "test":
            "http://yann.lecun.com/exdb/mnist/t10k-labels-idx1-ubyte.gz"
    }
}

_MNITS_PREPROCESSED = select_path(os.path.join(PREPROCESSED_BASE_DIR,
                                               'MNIST_preprocessed'),
                                  create_new=True)


def read_MNIST(override=False, verbose=False):
  download_path = os.path.join(DOWNLOAD_DIR, 'MNIST_original')
  if not os.path.exists(download_path):
    os.mkdir(download_path)

  if override:
    shutil.rmtree(_MNITS_PREPROCESSED)
    os.mkdir(_MNITS_PREPROCESSED)

  if len(os.listdir(_MNITS_PREPROCESSED)) == 0:
    # ====== load the values ====== #
    values = {}
    for kind, url in _URLs['values'].items():
      base_name = os.path.basename(url)
      zip_path = os.path.join(download_path, base_name)
      # download zip file
      if not os.path.exists(zip_path):
        get_file(fname=base_name,
                 origin=url,
                 outdir=download_path,
                 verbose=verbose)
      # extract the file
      with gzip.open(zip_path, "rb") as values_stream:
        _, M, r, c = struct.unpack(">IIII", values_stream.read(16))
        values_buffer = values_stream.read(M * r * c)
        values_flat = np.frombuffer(values_buffer, dtype=np.uint8)
        values[kind] = values_flat.reshape(-1, r * c)
    N = r * c

    # ====== load the labels ====== #
    labels = {}
    for kind, url in _URLs["labels"].items():
      base_name = os.path.basename(url)
      zip_path = os.path.join(download_path, base_name)
      # download zip file
      if not os.path.exists(zip_path):
        get_file(fname=base_name,
                 origin=url,
                 outdir=download_path,
                 verbose=verbose)
      # extract the file
      with gzip.open(zip_path, "rb") as labels_stream:
        _, M = struct.unpack(">II", labels_stream.read(8))
        labels_buffer = labels_stream.read(M)
        labels[kind] = np.frombuffer(labels_buffer, dtype=np.int8)

    # ====== post processing ====== #
    X_train, X_test = values["training"].astype(
        'float32'), values["test"].astype('float32')
    y_train, y_test = labels["training"], labels["test"]
    X_train = np.concatenate((X_train, X_test), axis=0)
    y_train = np.concatenate((y_train, y_test), axis=0)

    all_classes = [i for i in sorted(np.unique(y_train))]
    cls_2_idx = {c: i for i, c in enumerate(all_classes)}
    y_train = one_hot(np.array([cls_2_idx[i] for i in y_train]),
                      nb_classes=len(all_classes),
                      dtype='float32')
    all_classes = np.array(['#%d' % i for i in all_classes], dtype="U")

    X_train_row = np.array(
        ["image {}".format(i + 1) for i in range(X_train.shape[0])])

    feature_names = np.array(
        ["pixel {}".format(j + 1) for j in range(X_train.shape[1])])
    # ====== saving ====== #
    data_meta = {
        'X': X_train,
        'X_row': X_train_row,
        'X_col': feature_names,
        'y': y_train,
        'y_col': all_classes
    }
    for name, val in data_meta.items():
      path = os.path.join(_MNITS_PREPROCESSED, name)
      if verbose:
        print("Saving '%s' to path: %s" % (name, path))
      with open(path, 'wb') as f:
        pickle.dump(val, f)
  # ====== load the dataset ====== #
  ds = F.Dataset(_MNITS_PREPROCESSED, read_only=True)
  return ds



from __future__ import absolute_import, division, print_function

import os
import shutil

import numpy as np
from six import string_types

from odin.fuel import Dataset
from odin.utils import as_tuple
from sisua.data.data_loader.pbmc8k import read_PBMC8k
from sisua.data.data_loader.pbmcecc import read_PBMCeec
from sisua.data.path import PREPROCESSED_BASE_DIR
from sisua.data.utils import save_to_dataset, standardize_protein_name


def read_PBMCcross_ecc_8k(subset,
                          return_ecc,
                          filtered_genes=False,
                          override=False,
                          verbose=False):
  """ This create a dataset with shared genes subset between
  PBMC-ecc and PBMC-8k

  It will select the most overlap subset when `filtered_genes=True`
  """
  preprocessed_path = os.path.join(
      PREPROCESSED_BASE_DIR,
      'PBMCcross_%s_%s_preprocessed' % ('ecc' if return_ecc else '8k', subset +
                                        ('' if filtered_genes else 'full')))
  if override and os.path.exists(preprocessed_path):
    shutil.rmtree(preprocessed_path)
  if not os.path.exists(preprocessed_path):
    os.mkdir(preprocessed_path)

  # ******************** preprocessed ******************** #
  if not os.path.exists(os.path.join(preprocessed_path, 'X')):
    pbmc8k_full = read_PBMC8k(subset=subset,
                              override=override,
                              filtered_genes=False,
                              verbose=verbose)
    pbmcecc_full = read_PBMCeec(subset=subset,
                                override=override,
                                filtered_genes=False,
                                verbose=verbose)

    all_genes = set(pbmc8k_full['X_col']) & set(pbmcecc_full['X_col'])

    if filtered_genes:
      pbmc8k = read_PBMC8k(subset=subset,
                           override=override,
                           filtered_genes=True,
                           verbose=verbose)
      pbmcecc = read_PBMCeec(subset=subset,
                             override=override,
                             filtered_genes=True,
                             verbose=verbose)
      s1 = set(pbmc8k['X_col']) & all_genes
      s2 = set(pbmcecc['X_col']) & all_genes
      all_genes = s2 if len(s2) > len(s1) else s1
    # the same order all the time
    all_genes = sorted(all_genes)

    pbmc = pbmcecc_full if return_ecc else pbmc8k_full
    X = pbmc['X']
    X_row = pbmc['X_row']
    X_col = pbmc['X_col']
    y = pbmc['y']
    y_col = pbmc['y_col']

    X_col_indices = {gene: i for i, gene in enumerate(X_col)}
    indices = np.array([X_col_indices[gene] for gene in all_genes])
    X = X[:, indices]
    X_col = X_col[indices]
    save_to_dataset(preprocessed_path,
                    X,
                    X_col,
                    y,
                    y_col,
                    rowname=X_row,
                    print_log=verbose)
  # ******************** return ******************** #
  ds = Dataset(preprocessed_path, read_only=True)
  return ds


# ===========================================================================
# Remove set of protein
# ===========================================================================
def read_PBMCcross_remove_protein(subset,
                                  return_ecc,
                                  filtered_genes=False,
                                  override=False,
                                  verbose=False,
                                  remove_protein=['CD4', 'CD8']):
  remove_protein = sorted(
      [i.lower() for i in as_tuple(remove_protein, t=string_types)])
  preprocessed_path = os.path.join(
      PREPROCESSED_BASE_DIR, 'PBMCcross_%s_%s_no%s_preprocessed' %
      ('ecc' if return_ecc else '8k', subset +
       ('' if filtered_genes else 'full'), ''.join(
           [i.lower() for i in remove_protein])))
  if override and os.path.exists(preprocessed_path):
    shutil.rmtree(preprocessed_path)
  if not os.path.exists(preprocessed_path):
    os.mkdir(preprocessed_path)

  # ******************** preprocessed ******************** #
  if not os.path.exists(os.path.join(preprocessed_path, 'X')):
    ds = read_PBMCcross_ecc_8k(subset,
                               return_ecc,
                               filtered_genes,
                               override=override,
                               verbose=verbose)
    X = ds['X'][:]
    X_row = ds['X_row']
    X_col = ds['X_col']
    y = ds['y']
    y_col = ds['y_col']

    remove_ids = [
        i for i, j in enumerate(y_col)
        if standardize_protein_name(j).lower() in remove_protein
    ]
    remain_ids = [i for i in range(len(y_col)) if i not in remove_ids]
    y_col = y_col[remain_ids]
    y = y[:, remain_ids]

    save_to_dataset(preprocessed_path,
                    X,
                    X_col,
                    y,
                    y_col,
                    rowname=X_row,
                    print_log=verbose)
  # ******************** return ******************** #
  ds = Dataset(preprocessed_path, read_only=True)
  return ds



from __future__ import print_function, division, absolute_import

import io
from io import BytesIO
from collections import defaultdict

import numpy as np
from tqdm import tqdm

from odin.utils import batching, stdio, get_formatted_datetime

from .inference_base import Inference
from ._consts import PREDICTION_BATCH_SIZE
from sisua.data import get_library_size, apply_artificial_corruption
from sisua.data.const import UNIVERSAL_RANDOM_SEED

# ===========================================================================
# scVI Helpers
# ===========================================================================
try:
  import torch
  torch_device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
  torch.manual_seed(UNIVERSAL_RANDOM_SEED)
except ImportError:
  torch_device = None

def _create_gene_dataset(X, local_means=None, local_vars=None):
  from scvi.dataset import GeneExpressionDataset
  if local_means is None or local_vars is None:
    local_means, local_vars = get_library_size(X)
  col_names = np.array(['Gene#%d' % i for i in range(X.shape[1])])
  gene_dataset = GeneExpressionDataset(
      X=X,
      local_means=local_means,
      local_vars=local_vars,
      batch_indices=np.zeros(shape=(X.shape[0], 1)),
      labels=None,
      gene_names=col_names,
      cell_types=None)
  gene_dataset.labels = X
  gene_dataset.cell_types = col_names
  return gene_dataset

def _create_data_iter(X, log_variational):
  import torch
  # tqdm(,ncols=80, desc="Predicting")
  for start, end in batching(batch_size=PREDICTION_BATCH_SIZE // 3,
                             n=X.shape[0]):
    x = _to_tensor(X[start:end])
    if bool(log_variational):
      x = torch.log(1 + x)
    yield x

def _to_tensor(x):
  import torch
  return torch.from_numpy(x.astype('float32')).to(torch_device)

def _to_array(x):
  if hasattr(x, 'todense'):
    return np.array(x.todense())
  if hasattr(x, 'cpu'):
    return x.data.cpu().numpy()
  return x

def _use_cuda():
  import torch
  return torch.cuda.is_available()

# ===========================================================================
# Main
# ===========================================================================
class InferenceSCVI(Inference):
  """ InferenceSCVI """

  def __init__(self, gene_dim,
               dispersion='gene-cell',
               xnorm='log', tnorm='raw', ynorm='prob',
               xclip=0, yclip=0,
               xdist='zinb', ydist='bernoulli', zdist='normal',
               xdrop=0.3, edrop=0, zdrop=0, ddrop=0,
               hdim=128, zdim=32, nlayer=2,
               batchnorm=True, analytic=True,
               kl_weight=1., warmup=400, y_weight=10.,
               extra_module_path=None, no_library_size=False,
               **kwargs):
    try:
      import scvi
    except ImportError as e:
      raise RuntimeError(
          "scVI package at https://github.com/YosefLab/scVI is required")
    model = kwargs.get('model', 'vae')
    prot_dim = None
    if model not in ('vae', 'scvi'):
      raise RuntimeError("InferenceSCVI only support 'vae' model")
    if xdist not in ('zinb', 'nb'):
      raise ValueError("Only support 'zinb' and 'nb' for `xdist`")
    # ====== main model ====== #
    if xnorm not in ('log', 'raw'):
      raise ValueError(
          "Only support log variational or raw count variational for scVI")
    if tnorm not in ('raw',):
      raise ValueError(
          "Only support raw count target value for scVI")
    self._name = 'scvinl' if no_library_size else 'scvi'
    # ====== others ====== #
    self._history = {'train': defaultdict(list),
                     'valid': defaultdict(list)}
    self._training_log = ''
    self._configs = dict(locals())
    self._configs.update(kwargs)

    self._gmm_threshold = {}
    self._gene_dim = int(gene_dim)
    self._prot_dim = 0 if prot_dim is None else max(int(prot_dim), 0)
    self._is_fitted = False

    self._ps = 0
    self._trained_n_epoch = 0
    self._batch_size = 0
    self._learning_rate = 0
    self._n_mcmc_train = 0
    self._corruption_rate = None
    self._corruption_dist = None
    self._initialize_model()
    self._no_library_size = bool(no_library_size)

  @property
  def no_library_size(self):
    return self._no_library_size

  # ******************** pickling ******************** #
  def _initialize_model(self, weights=None):
    from scvi.models import VAE
    self._vae = VAE(n_input=self.gene_dim, n_batch=0, n_labels=0,
                    n_hidden=self.hdim, n_latent=self.zdim, n_layers=self.nlayer,
                    dropout_rate=self.xdrop,
                    dispersion=self.dispersion,
                    log_variational=True if self.xnorm == 'log' else False,
                    reconstruction_loss=self.xdist)
    if weights is not None:
      data = BytesIO(weights)
      self._vae.load_state_dict(torch.load(data))
      self._vae.to(torch_device)

  def __getstate__(self):
    import torch
    states = self.get_states()
    weights = BytesIO()
    torch.save(self._vae.state_dict(), weights)
    weights.seek(0)
    weights = weights.read()
    return states, weights, self._no_library_size

  def __setstate__(self, states):
    states, weights, self._no_library_size = states
    self.set_states(states)
    self._initialize_model(weights=weights)

  # ******************** fitting ******************** #
  def fit(self, X, y=None,
          supervised_percent=0.8, validation_percent=0.1,
          n_mcmc_samples=1,
          corruption_rate=0.25, corruption_dist='binomial',
          batch_size=64, n_epoch=120, learning_rate=1e-4,
          monitoring=False, fig_dir=None,
          detail_logging=False):
    training_log = io.StringIO()
    stdio(training_log)

    if self._corruption_rate is None:
      self._corruption_rate = corruption_rate
      self._corruption_dist = corruption_dist
    else:
      assert (self._corruption_rate == corruption_rate and
              self._corruption_dist == corruption_dist),\
      ("Model have been fitted with following corruption configurations: %s, %s;"
       "but given: %s, %s" % (self._corruption_rate, self._corruption_dist,
                              corruption_rate, corruption_dist))
    assert n_mcmc_samples == 1, \
    "Only support 1 MCMC sample during training for scVI"

    self._ps = supervised_percent
    self._batch_size = batch_size
    self._learning_rate = learning_rate
    self._n_mcmc_train = n_mcmc_samples
    X = apply_artificial_corruption(
        X, dropout=corruption_rate, distribution=corruption_dist)
    # ====== check input dims ====== #
    assert X.shape[1] == self.gene_dim,\
    "Number of gene expression mismatch: %d != %d" % (self.gene_dim, X.shape[1])
    if validation_percent == 0:
      raise RuntimeError("scVI doesn't support 0% validation percent!")
    # ====== training ====== #
    from scvi.inference import UnsupervisedTrainer
    gene_dataset = _create_gene_dataset(X)
    # batchnorm bug in scVI, if return batch is 1, then
    # it raise Exception, just modify the batch size so
    # it never return any batch with size 1 for Cortex dataset
    # wouldn't make a significant different
    if 'dataset' in self.configs and self.configs['dataset'] == 'cortex':
      batch_size = batch_size - 1
    trainer = UnsupervisedTrainer(self._vae, gene_dataset,
        use_cuda=_use_cuda(),
        train_size=1 - validation_percent,
        frequency=1, benchmark=False,
        data_loader_kwargs=dict(batch_size=batch_size),
        metrics_to_monitor=None,
        verbose=detail_logging)
    # scVI will do the normalization by itself
    trainer.train(n_epochs=n_epoch, lr=learning_rate)
    self._scvi_trainer = trainer
    # ====== save history ====== #
    self._history['train']['loss'] += trainer.history["ll_train_set"][1:]
    self._history['valid']['loss'] += trainer.history["ll_test_set"][1:]
    # ====== update states ====== #
    self._trained_n_epoch += n_epoch
    self._is_fitted = True
    stdio()
    training_log.seek(0)
    self._training_log += '\n%s\n' % get_formatted_datetime(only_number=False)
    self._training_log += training_log.read()
    training_log.close()

  # ******************** history ******************** #
  def get_train_loss(self):
    return self._history['train']['loss']

  def get_valid_loss(self):
    return self._history['valid']['loss']

  def get_train_history(self, name):
    return self._history['train'][name]

  def get_valid_history(self, name):
    return self._history['valid'][name]

  # ******************** scoring ******************** #
  def score(self, X, y=None, n_mcmc_samples=100, return_mean=True):
    """Compute the per-sample average of recorded metrics of the given data X.

    Parameters
    ----------
    X : array-like, shape (n_samples, n_dimensions)
        List of n_features-dimensional data points. Each row
        corresponds to a single data point.

    Returns
    -------
    dictionary : metric_name -> values
    """
    # only VAE model for scVI
    self._vae.eval()
    from scvi.inference import Posterior
    pos = Posterior(model=self._vae, gene_dataset=_create_gene_dataset(X),
                    shuffle=False, use_cuda=_use_cuda(),
                    data_loader_kwargs=dict(batch_size=PREDICTION_BATCH_SIZE // 3))
    ll = pos.ll(verbose=False)
    marginal_ll = pos.marginal_ll(verbose=False, n_mc_samples=n_mcmc_samples)
    return {'loss': ll, 'llk': marginal_ll}

  # ******************** predicting ******************** #
  def predict_Z(self, X):
    """ Return mean of the latent posterior
    (i.e. mean of Normal distribution) """
    self._vae.eval()
    from scvi.inference import Posterior
    pos = Posterior(model=self._vae, gene_dataset=_create_gene_dataset(X),
                    shuffle=False, use_cuda=_use_cuda(),
                    data_loader_kwargs={'batch_size': 64})
    return pos.get_latent(sample=False)[0]

  def predict_y(self, X, y, n_mcmc_samples=100):
    """ Predicting the protein marker """
    assert self.is_semi_supervised and y is not None,\
    "Only semi-supervised model has this prediction, and `y` must not be None"

  def predict_L(self, X, n_mcmc_samples=100):
    """ Return library size prediction """
    self._vae.eval()
    L = []
    for x in _create_data_iter(X, log_variational=self._vae.log_variational):
      ql_m, ql_v, library = self._vae.l_encoder(x)
      L.append(_to_array(ql_m))
    return np.concatenate(L, axis=0)

  def predict_Lstddev(self, X, n_mcmc_samples=100):
    """ Return library size stddev prediction """
    self._vae.eval()
    Lstddev = []
    for x in _create_data_iter(X, log_variational=self._vae.log_variational):
      ql_m, ql_v, library = self._vae.l_encoder(x)
      Lstddev.append(_to_array(ql_v))
    return np.concatenate(Lstddev, axis=0)

  def predict_W(self, X, n_mcmc_samples=100):
    """ Return mean of the reconstructed value """
    self._vae.eval()
    W = []
    for x in _create_data_iter(X, log_variational=False):
      # scVI do log normalization within .inference()
      (px_scale, px_r, px_rate, px_dropout,
       qz_m, qz_v, z,
       ql_m, ql_v, library) = self._vae.inference(
          x, batch_index=None, y=None, n_samples=n_mcmc_samples)
      if self.no_library_size:
        px_rate = px_scale * x.sum(1).unsqueeze(1)
      w = (1 - px_dropout) * px_rate
      w = w.mean(0)
      W.append(_to_array(w))
    return np.concatenate(W, axis=0)

  def predict_Wstddev(self, X, n_mcmc_samples=100):
    """ Return stddev of reconstructed value
    if not a variational model, then V_stddev is None
    """
    self._vae.eval()
    import torch
    Wstddev = []
    for x in _create_data_iter(X, log_variational=False):
      # scVI do log normalization within .inference()
      (px_scale, px_r, px_rate, px_dropout,
       qz_m, qz_v, z,
       ql_m, ql_v, library) = self._vae.inference(
          x, batch_index=None, y=None, n_samples=n_mcmc_samples)
      w_mean = (1 - px_dropout) * px_rate
      # (1 - pi) * (d.var + d.mean^2) - [(1 - pi) * d.mean]^2
      w_var = (1 - px_dropout) * (px_r + px_rate**2) - w_mean**2
      w_std = torch.sqrt(w_var).mean(0)
      Wstddev.append(_to_array(w_std))
    return np.concatenate(Wstddev, axis=0)

  def predict_V(self, X, n_mcmc_samples=100):
    """ Return mean of the denoised value """
    self._vae.eval()
    V = []
    for x in _create_data_iter(X, log_variational=False):
      # scVI do log normalization within .inference()
      if self._no_library_size:
        out = self._vae.get_sample_scale(x, n_samples=n_mcmc_samples).mean(0)
        out = out * x.sum(1).unsqueeze(1)
      else:
        out = self._vae.get_sample_rate(x, n_samples=n_mcmc_samples).mean(0)
      V.append(_to_array(out))
    return np.concatenate(V, axis=0)

  def predict_Vstddev(self, X, n_mcmc_samples=100):
    """ Return stddev of denoised value
    if not a variational model, then V_stddev is None
    """
    self._vae.eval()
    Vstddev = []
    for x in _create_data_iter(X, log_variational=False):
      n = x.shape[0]
      # px_r
      out = self._vae.inference(
          x, batch_index=None, y=None, n_samples=n_mcmc_samples)[1]
      if self.dispersion == 'gene-cell':
        out = np.mean(_to_array(out), axis=0)
      else:
        out = np.expand_dims(_to_array(out), axis=0)
        out = np.repeat(out, repeats=n, axis=0)
      Vstddev.append(out)
    return np.concatenate(Vstddev, axis=0)

  def predict_PI(self, X, n_mcmc_samples=100):
    """ Return a matrix (n_sample, n_gene) of Zero-inflated
    rate

    if not a zero-inflated model, then return None
    """
    self._vae.eval()
    import torch
    op = torch.nn.Sigmoid()
    PI = []
    for x in _create_data_iter(X, log_variational=False):
      # px_dropout (logits)
      out = self._vae.inference(
          x, batch_index=None, y=None, n_samples=n_mcmc_samples)[3]
      out = op(out.mean(0))
      out = _to_array(out)
      PI.append(out)
    return np.concatenate(PI, axis=0)

  # ******************** sampling function ******************** #
  def sample_Z(self, X, n_mcmc_samples=1):
    self._vae.eval()
    from scvi.inference import Posterior
    pos = Posterior(model=self._vae, gene_dataset=_create_gene_dataset(X),
                    shuffle=False, use_cuda=_use_cuda(),
                    data_loader_kwargs={'batch_size': 64})

    return np.concatenate(
        [np.expand_dims(pos.get_latent(sample=True)[0], axis=0)
         for i in range(n_mcmc_samples)], axis=0)

  def sample_y(self, X, n_mcmc_samples=1):
    raise NotImplementedError

  def sample_W(self, X, n_mcmc_samples=1):
    raise NotImplementedError

  def sample_V(self, X, n_mcmc_samples=1):
    self._vae.eval()
    V = []
    for x in _create_data_iter(X, log_variational=False):
      # scVI do log normalization within .inference()
      out = self._vae.get_sample_rate(x, n_samples=n_mcmc_samples)
      V.append(_to_array(out))
    return np.concatenate(V, axis=1)

  def sample_L(self, X, n_mcmc_samples=1):
    raise NotImplementedError



from __future__ import absolute_import, division, print_function

import os
import warnings
from copy import deepcopy
from typing import Optional, Tuple

import numpy as np
import scanpy as sc
import scipy as sp
import tensorflow as tf
from anndata.core.alignedmapping import AxisArrays
from scipy.sparse import issparse
from six import string_types

from odin import visual as vs
from odin.fuel import MmapArrayWriter
from odin.stats import describe, sparsity_percentage, train_valid_test_split
from odin.utils import (as_tuple, batching, cache_memory, catch_warnings_ignore,
                        ctext)
from odin.utils.crypto import md5_checksum
from odin.visual import Visualizer, to_axis
from sisua.label_threshold import ProbabilisticEmbedding

# ===========================================================================
# Heuristic constants
# ===========================================================================
_DEFAULT_BATCH_SIZE = 4096


# ===========================================================================
# Helper
# ===========================================================================
def apply_artificial_corruption(x, dropout, distribution, copy=False, seed=8):
  """
  Parameters
  ----------
  x : (n_samples, n_features)
  dropout : scalar (0.0 - 1.0)
  distribution : {'uniform', 'binomial
  """
  distribution = str(distribution).lower()
  dropout = float(dropout)
  assert 0 <= dropout < 1, \
  "dropout value must be >= 0 and < 1, given: %f" % dropout
  rand = np.random.RandomState(seed=seed)

  if dropout == 0:
    return x
  # ====== applying corruption ====== #
  # Original code from scVI, to provide a comparable result,
  # please acknowledge the author of scVI if you are using this
  # code for corrupting the data
  # https://github.com/YosefLab/scVI/blob/2357dde15351450e452efa426c516c60a2d5ee96/scvi/dataset/dataset.py#L83
  # the test data won't be corrupted
  if copy:
    corrupted_x = deepcopy(x)
  else:
    corrupted_x = x

  # multiply the entry n with a Ber(0.9) random variable.
  if distribution == "uniform":
    i, j = np.nonzero(x)
    ix = rand.choice(range(len(i)),
                     size=int(np.floor(dropout * len(i))),
                     replace=False)
    i, j = i[ix], j[ix]
    corrupted = np.multiply(
        x[i, j], rand.binomial(n=np.ones(len(ix), dtype=np.int32), p=0.9))
  # multiply the entry n with a Bin(n, 0.9) random variable.
  elif distribution == "binomial":
    i, j = (k.ravel() for k in np.indices(x.shape))
    ix = rand.choice(range(len(i)),
                     size=int(np.floor(dropout * len(i))),
                     replace=False)
    i, j = i[ix], j[ix]
    # only 20% expression captured
    corrupted = rand.binomial(n=(x[i, j]).astype(np.int32), p=0.2)
  else:
    raise ValueError(
        "Only support 2 corruption distribution: 'uniform' and 'binomial', "
        "but given: '%s'" % distribution)

  if isinstance(corrupted_x, sp.sparse.base.spmatrix):
    corrupted = type(corrupted_x)(corrupted)
  corrupted_x[i, j] = corrupted
  return corrupted_x


def get_library_size(X, return_log_count=False):
  """ Copyright scVI authors
  https://github.com/YosefLab/scVI/blob/master/README.rst

  Original Code:
  https://github.com/YosefLab/scVI/blob/9d9a525df810c47ce482ef7b554f25fcc6482c2d/scvi/dataset/dataset.py#L288

  size factor of X in log-space

  Parameters
  ----------
  X : matrix
    single-cell data matrix (n_samples, n_features)
  return_log_count : bool (default=False)
    if True, return the log-count library size

  Return
  ------
  local_mean (n_samples, 1)
  local_var (n_samples, 1)
  """
  assert X.ndim == 2, "Only support 2-D matrix"
  total_counts = X.sum(axis=1)
  assert np.all(total_counts >= 0), "Some cell contains negative-count!"
  log_counts = np.log(total_counts + 1e-8)
  local_mean = (np.mean(log_counts) * np.ones(
      (X.shape[0], 1))).astype(np.float32)
  local_var = (np.var(log_counts) * np.ones((X.shape[0], 1))).astype(np.float32)
  if not return_log_count:
    return local_mean, local_var
  return np.expand_dims(log_counts, -1), local_mean, local_var


# ===========================================================================
# OMICS
# ===========================================================================
class SingleCellOMIC(sc.AnnData, Visualizer):
  """ An annotated data matrix.

  Parameters
  ----------
  X
      A #observations × #variables data matrix. A view of the data is used
      if the data type matches, otherwise, a copy is made.
  obs
      Key-indexed one-dimensional observations annotation of length
      #observations.
  var
      Key-indexed one-dimensional variables annotation of length #variables.
  uns
      Key-index unstructured annotation.
  obsm
      Key-indexed multi-dimensional observations annotation of length
      #observations. If passing a :class:`~numpy.ndarray`,
      it needs to have a structured datatype.
  varm
      Key-indexed multi-dimensional variables annotation of length #variables.
      If passing a :class:`~numpy.ndarray`, it needs to have a structured
      datatype.
  dtype
      Data type used for storage.
  shape
      Shape tuple (#observations, #variables). Can only be provided
      if ``X`` is ``None``.
  filename
      Name of backing file. See :class:`anndata.h5py.File`.
  filemode
      Open mode of backing file. See :class:`anndata.h5py.File`.
  layers
      Dictionary with keys as layers' names and values as matrices of the
      same dimensions as X.
  """

  def __init__(self,
               X=None,
               obs=None,
               var=None,
               uns=None,
               obsm=None,
               varm=None,
               layers=None,
               raw=None,
               dtype='float32',
               shape=None,
               filename=None,
               filemode=None,
               asview=False,
               oidx=None,
               vidx=None,
               name="scOMICS"):
    if X is not None:
      if obs is None:
        obs = {'rowid': ['Row#%d' % i for i in range(X.shape[0])]}
      if var is None:
        var = {'colid': ['Col#%d' % i for i in range(X.shape[1])]}
    super(SingleCellOMIC, self).__init__(X=X,
                                         obs=obs,
                                         var=var,
                                         uns=uns,
                                         obsm=obsm,
                                         varm=varm,
                                         layers=layers,
                                         raw=raw,
                                         dtype=dtype,
                                         shape=shape,
                                         filename=filename,
                                         filemode=filemode,
                                         asview=asview,
                                         oidx=oidx,
                                         vidx=vidx)
    self._name = str(name)
    self._indices = np.arange(self.X.shape[0], dtype='int32')
    self._calculate_library_info()

  @property
  def X(self):
    with catch_warnings_ignore(FutureWarning):
      X = super(SingleCellOMIC, self).X
    if X.ndim == 1:
      X = np.expand_dims(X, axis=1)
    return X

  def assert_matching_cells(self, sco) -> 'SingleCellOMIC':
    assert isinstance(sco, SingleCellOMIC), \
      "sco must be instance of SingleCellOMIC"
    assert sco.shape[0] == self.shape[0], \
      "Number of cell mismatch %d and %d" % (self.shape[0], sco.shape[0])
    if 'cellid' in sco.obs and 'cellid' in self.obs:
      assert np.all(sco.obs['cellid'] == self.obs['cellid'])
    else:  # just check matching first column
      assert np.all(sco.obs.iloc[:, 0] == self.obs.iloc[:, 0])
    return self

  def _calculate_library_info(self):
    if sp.sparse.issparse(self.X):
      total_counts = np.expand_dims(np.sum(self.X, axis=1), axis=-1)
    else:
      total_counts = np.sum(self.X, axis=1, keepdims=True)

    log_counts, local_mean, local_var = get_library_size(self.X,
                                                         return_log_count=True)
    self._total_counts = total_counts
    self._log_counts = log_counts
    self._local_mean = local_mean
    self._local_var = local_var

  def as_obsm(self, obsm_name):
    assert obsm_name in self.obsm
    omics = SingleCellOMIC(self,
                           oidx=slice(None, None),
                           vidx=slice(None, None),
                           asview=True)
    X_org = omics._X
    omics._X = omics.obsm[obsm_name]
    # already applied obsm at least once
    if hasattr(self, '_obsm_name'):
      omics.obsm[self._obsm_name] = X_org
    else:
      omics.obsm['X_org'] = X_org
    # update the tracking information
    del omics.obsm[obsm_name]
    omics._obsm_name = obsm_name
    return omics

  def __getitem__(self, index):
    """Returns a sliced view of the object."""
    oidx, vidx = self._normalize_indices(index)
    omics = SingleCellOMIC(self, oidx=oidx, vidx=vidx, asview=True)
    # update observation indexing
    omics._obs = self._obs.iloc[oidx]
    omics._obsm = AxisArrays(omics,
                             0,
                             vals={i: j[oidx] for i, j in self._obsm.items()})
    omics._indices = self._indices[oidx]
    # update variable indexing
    omics._var = self._var.iloc[vidx]
    omics._varm = AxisArrays(omics,
                             1,
                             vals={i: j[vidx] for i, j in self._varm.items()})
    # other meta
    omics._n_obs = omics._obs.shape[0]
    omics._n_vars = omics._var.shape[0]
    omics._name = self._name + '_index'
    # library info
    omics._total_counts = self._total_counts[oidx]
    omics._log_counts = self._log_counts[oidx]
    omics._local_mean = self._local_mean[oidx]
    omics._local_var = self._local_var[oidx]
    return omics

  def apply_indices(self, indices, observation=True):
    """ Inplace indexing, this indexing algorithm also update
    `obs`, `obsm`, `var`, `varm` to complement with the new indices.

    Parameters
    ----------
    indices : array of `int` or `bool`
    observation : `bool` (default=True)
      if True, applying the indices to the observation (i.e. axis=0),
      otherwise, to the variable (i.e. axis=1)
    """
    indices = np.array(indices)
    itype = indices.dtype.type
    if not issubclass(itype, (np.bool, np.bool_, np.integer)):
      raise ValueError("indices type must be boolean or integer.")
    if observation:
      self._X = self._X[indices]
      self._obs = self._obs.iloc[indices]
      self._n_obs = self._X.shape[0]
      self._obsm = AxisArrays(
          self, 0, vals={i: j[indices] for i, j in self._obsm.items()})
      self._indices = self._indices[indices]
      self._total_counts = self._total_counts[indices]
      self._log_counts = self._log_counts[indices]
      self._local_mean = self._local_mean[indices]
      self._local_var = self._local_var[indices]
    else:
      self._X = self._X[:, indices]
      self._var = self._var.iloc[indices]
      self._n_vars = self._X.shape[1]
      self._varm = AxisArrays(
          self, 1, vals={i: j[indices] for i, j in self._varm.items()})
    return self

  def copy(self, X=None, filename=None, name=None):
    """Full copy, optionally on disk. (this code is copied from
    `AnnData`, modification to return `SingleCellOMIC` instance.
    """
    from anndata.core.views import DictView
    if not self.isbacked:
      omics = SingleCellOMIC(
          X if X is not None else
          (self._X.copy() if self._X is not None else None),
          self._obs.copy(),
          self._var.copy(),
          # deepcopy on DictView does not work and is unnecessary
          # as uns was copied already before
          self._uns.copy()
          if isinstance(self._uns, DictView) else deepcopy(self._uns),
          self._obsm.copy(),
          self._varm.copy(),
          raw=None if self._raw is None else self._raw.copy(),
          layers=dict(self.layers),
          dtype=self._X.dtype.name if self._X is not None else 'float32',
          name=self.name + '_copy' if name is None else name)
    else:
      if filename is None:
        raise ValueError(
            'To copy an SingleCellOMIC object in backed mode, '
            'pass a filename: `.copy(filename=\'myfilename.h5ad\')`.')
      if self.isview:
        self.write(filename)
      else:
        from shutil import copyfile
        copyfile(self.filename, filename)
      omics = SingleCellOMIC(filename=filename,
                             name=self.name if name is None else name)
    # other info related to SingleCellOMIC
    omics._indices = self.indices
    omics._total_counts = self._total_counts
    omics._log_counts = self._log_counts
    omics._local_mean = self._local_mean
    omics._local_var = self._local_var
    return omics

  @property
  def name(self):
    return self._name

  @property
  def is_binary(self):
    """ return True if the data is binary """
    return sorted(np.unique(self.X.astype('float32'))) == [0., 1.]

  @property
  def indices(self):
    """ Return the row indices had been used to created this data,
    helpful when using `SingleCellOMIC.split` to keep track the
    data partition """
    return self._indices

  @property
  def n_obs(self):
    """Number of observations."""
    return self._n_obs if self._X is None else self._X.shape[0]

  @property
  def n_vars(self):
    """Number of variables/features."""
    return self._n_vars if self._X is None else self._X.shape[1]

  @property
  def dtype(self):
    return self.X.dtype

  @property
  def library_size(self):
    """ Return the mean and variance for library size
    modeling in log-space """
    return self._local_mean, self._local_var

  @property
  def local_mean(self):
    return self._local_mean

  @property
  def local_var(self):
    return self._local_var

  @property
  def total_counts(self):
    return self._total_counts

  @property
  def log_counts(self):
    return self._log_counts

  # ******************** converter ******************** #
  def as_tensorflow_dataset(self, obs=(), obsm=(),
                            include_x=True) -> tf.data.Dataset:
    data = [self.X] if include_x else []
    for name in as_tuple(obs, t=string_types):
      data.append(self.obs[name].values)
    for name in as_tuple(obsm, t=string_types):
      data.append(self.obsm[name])
    assert len(data) > 0, "No data is given"
    ds = tf.data.Dataset.from_tensor_slices(data[0] if len(data) ==
                                            1 else tuple(data))
    return ds

  # ******************** transformation ******************** #
  def split(self, seed=8, train_percent=0.8) -> Tuple['SingleCellOMIC',
                                                      'SingleCellOMIC']:
    """ Spliting the data into training and test dataset

    Parameters
    ----------
    seed : `int` (default=8)
      the same seed will ensure the same partition of any `SingleCellOMIC`,
      as long as all the data has the same number of `SingleCellOMIC.nsamples`
    train_percent : `float` (default=0.8)
      the percent of data used for training, the rest is for testing
    add_name : `bool` (default=True)
      annotation the name of new data with 'train' and 'test'

    Returns
    -------
    train : `SingleCellOMIC`
    test : `SingleCellOMIC`

    Example
    -------
    >>> x_train, x_test = x.split()
    >>> y_train, y_test = y.split()
    >>> assert np.all(x_train.obs['cellid'] == y_train.obs['cellid'])
    >>> assert np.all(x_test.obs['cellid'] == y_test.obs['cellid'])
    >>> #
    >>> x_train_train, x_train_test = x_train.split()
    >>> assert np.all(x_train_train.obs['cellid'] ==
    >>>               y_train[x_train_train.indices].obs['cellid'])
    """
    assert 0 < train_percent < 1
    ids = np.random.RandomState(seed=seed).permutation(
        self.n_obs).astype('int32')
    ntrain = int(train_percent * self.n_obs)

    train_ids = ids[:ntrain]
    train = SingleCellOMIC(
        X=self.X[train_ids],
        obs=self.obs.iloc[train_ids],
        obsm={i: j[train_ids] for i, j in self.obsm.items()},
        var=self.var,
        varm=self.varm,
        uns=self.uns,  # it is tricky to split unstructed annotation
        name=self.name + '_train')
    train._indices = train_ids  # copy the indices, this is important

    test_ids = ids[ntrain:]
    test = SingleCellOMIC(
        X=self.X[test_ids],
        obs=self.obs.iloc[test_ids],
        obsm={i: j[test_ids] for i, j in self.obsm.items()},
        var=self.var,
        varm=self.varm,
        uns=self.uns,  # it is tricky to split unstructed annotation
        name=self.name + '_test')
    test._indices = test_ids  # copy the indices, this is important
    return train, test

  def corrupt(self,
              corruption_rate=0.25,
              corruption_dist='binomial',
              inplace=True,
              seed=8):
    """
    Parameters
    ----------
    corruption_rate : `float` (default=0.25)
    corruption_dist : {'binomial', 'uniform} (default='binomial')
    inplace : `bool` (default=True)
      Perform computation inplace or return new `SingleCellOMIC` with
      the corrupted data.
    seed : `int` (default=8)
        seed for the random state.

    """
    if corruption_rate <= 0:
      return self if inplace else self.copy()

    data = apply_artificial_corruption(self.X,
                                       corruption_rate,
                                       corruption_dist,
                                       copy=not inplace,
                                       seed=seed)
    name = '%s_%s%s' % (self.name, corruption_dist,
                        str(corruption_rate).split('.')[-1])
    if not inplace:
      omics = self.copy(X=data)
    else:
      omics = self
    omics._name = name
    omics._calculate_library_info()
    return omics

  def filter_highly_variable_genes(self,
                                   min_disp=0.5,
                                   max_disp=np.inf,
                                   min_mean=0.0125,
                                   max_mean=3,
                                   n_top_genes=1000,
                                   n_bins=20,
                                   flavor='cell_ranger',
                                   inplace=True):
    """ Annotate highly variable genes [Satija15]_ [Zheng17]_.

    Expects logarithmized data.

    Depending on `flavor`, this reproduces the R-implementations of Seurat
    [Satija15]_ and Cell Ranger [Zheng17]_.

    The normalized dispersion is obtained by scaling with the mean and standard
    deviation of the dispersions for genes falling into a given bin for mean
    expression of genes. This means that for each bin of mean expression, highly
    variable genes are selected.

    Parameters
    ----------
    min_disp : `float`, optional (default=0.5)
        If `n_top_genes` unequals `None`, this and all other cutoffs for
        the means and the normalized dispersions are ignored.
    max_disp : `float`, optional (default=`np.inf`)
        If `n_top_genes` unequals `None`, this and all other cutoffs for
        the means and the normalized dispersions are ignored.
    min_mean : `float`, optional (default=0.0125)
        If `n_top_genes` unequals `None`, this and all other cutoffs for
        the means and the normalized dispersions are ignored.
    max_mean : `float`, optional (default=3)
        If `n_top_genes` unequals `None`, this and all other cutoffs for
        the means and the normalized dispersions are ignored.
    n_top_genes : {`float`, int`, `None`}, optional (default=`None`)
        Number of highly-variable genes to keep., if the value is in (0, 1],
        intepret as percent of genes
    n_bins : `int`, optional (default: 20)
        Number of bins for binning the mean gene expression. Normalization is
        done with respect to each bin. If just a single gene falls into a bin,
        the normalized dispersion is artificially set to 1.
    flavor : `{'seurat', 'cell_ranger'}`, optional (default='seurat')
        Choose the flavor for computing normalized dispersion. In their default
        workflows, Seurat passes the cutoffs whereas Cell Ranger passes
        `n_top_genes`.
    inplace : `bool` (default=True)
        if False, copy the `SingleCellOMIC` and apply the vargene filter.

    Returns
    -------
    New `SingleCellOMIC` with filtered features if `applying_filter=True` else
    assign `SingleCellOMIC.highly_variable_features` with following attributes

    highly_variable : bool
        boolean indicator of highly-variable genes
    **means**
        means per gene
    **dispersions**
        dispersions per gene
    **dispersions_norm**
        normalized dispersions per gene

    Notes
    -----
    Proxy to `scanpy.pp.highly_variable_genes`
    It is recommended to do `log1p` normalization before if `flavor='seurat'`
    """
    flavor = str(flavor).lower()
    if n_top_genes is not None:
      if 0. < n_top_genes < 1.:
        n_top_genes = int(n_top_genes * self.n_vars)
      n_top_genes += 1
    # prepare the data
    # this function will take the exponential of X all the time,
    # so non-logarithmzed data might led to overflow
    omics = self if inplace else self.copy()
    sc.pp.highly_variable_genes(omics,
                                min_disp=min_disp,
                                max_disp=max_disp,
                                min_mean=min_mean,
                                max_mean=max_mean,
                                n_top_genes=n_top_genes,
                                flavor=flavor,
                                subset=True,
                                inplace=False)
    omics._name += '_vargene'
    omics._n_vars = omics._X.shape[1]
    # recalculate library info
    omics._calculate_library_info()
    return omics

  def filter_genes(self,
                   min_counts=None,
                   max_counts=None,
                   min_cells=None,
                   max_cells=None,
                   inplace=True):
    """ Filter features (columns) based on number of rows or counts.

    Keep columns that have at least ``[min_counts, max_counts]``
    or are expressed in at least ``[min_row_counts, max_row_counts]``

    Parameters
    ----------
    min_counts : {int, None} (default=None)
      Minimum number of counts required for a feature to pass filtering.
    max_counts : {int, None} (default=None)
      Maximum number of counts required for a feature to pass filtering.
    min_row_counts : {int, None} (default=None)
      Minimum number of rows expressed required for a feature to pass filtering.
    max_row_counts : {int, None} (default=None)
      Maximum number of rows expressed required for a feature to pass filtering.
    inplace : `bool` (default=True)
      if False, return new `SingleCellOMIC` with the filtered
      genes applied

    Returns
    -------
    if `applying_filter=False` annotates the `SingleCellOMIC`, otherwise,
    return new `SingleCellOMIC` with the new subset of genes

    gene_subset : `numpy.ndarray`
        Boolean index mask that does filtering. `True` means that the
        gene is kept. `False` means the gene is removed.
    number_per_gene : `numpy.ndarray`
        Depending on what was thresholded (`counts` or `cells`), the array
        stores `n_counts` or `n_cells` per gene.

    Note
    ----
    Proxy method to Scanpy preprocessing
    """
    omics = self if inplace else self.copy()
    sc.pp.filter_genes(omics,
                       min_counts=min_counts,
                       max_counts=max_counts,
                       min_cells=min_cells,
                       max_cells=max_cells,
                       inplace=True)
    omics._name += '_filtergene'
    omics._n_vars = omics._X.shape[1]
    # recalculate library info
    omics._calculate_library_info()
    return omics

  def filter_cells(self,
                   min_counts=None,
                   max_counts=None,
                   min_genes=None,
                   max_genes=None,
                   inplace=True):
    """ Filter examples (rows) based on number of features or counts.

    Keep rows that have at least ``[min_counts, max_counts]``
    or are expressed in at least ``[min_col_counts, max_col_counts]``

    Parameters
    ----------
    min_counts : {int, None} (default=None)
      Minimum number of counts required for a feature to pass filtering.
    max_counts : {int, None} (default=None)
      Maximum number of counts required for a feature to pass filtering.
    min_genes : {int, None} (default=None)
      Minimum number of rows expressed required for a feature to pass filtering.
    max_genes : {int, None} (default=None)
      Maximum number of rows expressed required for a feature to pass filtering.
    inplace : `bool` (default=True)
      if False, return new `SingleCellOMIC` with the filtered
      cells applied

    Returns
    -------
    if `applying_filter=False` annotates the `SingleCellOMIC`, otherwise,
    return new `SingleCellOMIC` with the new subset of cells

    cells_subset : numpy.ndarray
        Boolean index mask that does filtering. ``True`` means that the
        cell is kept. ``False`` means the cell is removed.
    number_per_cell : numpy.ndarray
        Depending on what was tresholded (``counts`` or ``genes``), the array stores
        ``n_counts`` or ``n_cells`` per gene.

    Note
    ----
    Proxy method to Scanpy preprocessing
    """
    # scanpy messed up here, the obs was not updated with the new indices
    cells_subset, number_per_cell = sc.pp.filter_cells(self,
                                                       min_counts=min_counts,
                                                       max_counts=max_counts,
                                                       min_genes=min_genes,
                                                       max_genes=max_genes,
                                                       inplace=False)
    omics = self if inplace else self.copy()
    omics.apply_indices(cells_subset, observation=True)
    omics._name += '_filtercell'
    return omics

  def probabilistic_embedding(self,
                              n_components_per_class=2,
                              positive_component=1,
                              log_norm=False,
                              clip_quartile=0.,
                              remove_zeros=True,
                              ci_threshold=-0.68,
                              seed=8,
                              pbe: Optional[ProbabilisticEmbedding] = None):
    """ Fit a GMM on each feature column to get the probability or binary
    representation of the features

    Parameters
    ----------
    pbe : {`sisua.ProbabilisticEmbedding`, `None`}
      pretrained instance of `ProbabilisticEmbedding` if given
    """
    # We turn-off default log_norm here since the data can be normalized
    # separately in advance.
    if self.shape[1] >= 100:
      warnings.warn("%d GMM will be trained!" % self.shape[1])

    if pbe is None:
      if 'pbe' not in self.uns:
        pbe = ProbabilisticEmbedding(
            n_components_per_class=n_components_per_class,
            positive_component=positive_component,
            log_norm=log_norm,
            clip_quartile=clip_quartile,
            remove_zeros=remove_zeros,
            ci_threshold=ci_threshold,
            random_state=seed)
        from sklearn.exceptions import ConvergenceWarning
        with catch_warnings_ignore(ConvergenceWarning):
          pbe.fit(self.X)
        self.uns['pbe'] = pbe
      else:
        pbe = self.uns['pbe']
    else:
      assert isinstance(pbe, ProbabilisticEmbedding), \
        'pbe, if given, must be instance of sisua.ProbabilisticEmbedding'
    self.obsm['X_prob'] = pbe.predict_proba(self._X)
    self.obsm['X_bin'] = pbe.predict(self._X)
    self._name += '_pbe'
    return self

  def pca(self, n_comps=50):
    """ Code from `scanpy.pp.pca`, there is bug when `chunked=True` so we
    modify it here. """
    from sklearn.decomposition import IncrementalPCA

    if self.n_vars < n_comps:
      n_comps = self.n_vars - 1

    X_pca = np.zeros((self.n_obs, n_comps), self.dtype)
    pca_ = IncrementalPCA(n_components=n_comps)

    for chunk, _, _ in self.chunked_X(_DEFAULT_BATCH_SIZE):
      chunk = chunk.toarray() if issparse(chunk) else chunk
      pca_.partial_fit(chunk)

    for chunk, start, end in self.chunked_X(_DEFAULT_BATCH_SIZE):
      chunk = chunk.toarray() if issparse(chunk) else chunk
      X_pca[start:end] = pca_.transform(chunk)

    self.obsm['X_pca'] = X_pca
    self.varm['PCs'] = pca_.components_.T
    self.uns['pca'] = {}
    self.uns['pca']['variance'] = pca_.explained_variance_
    self.uns['pca']['variance_ratio'] = pca_.explained_variance_ratio_
    return self

  def expm1(self, inplace=True):
    omics = self if inplace else self.copy()
    _expm1 = lambda x: (np.expm1(x.data, out=x.data)
                        if issparse(x) else np.expm1(x, out=x))
    for s, e in batching(n=self.n_obs, batch_size=_DEFAULT_BATCH_SIZE):
      omics._X[s:e] = _expm1(omics._X[s:e])
    omics._name += '_expm1'
    return omics

  def normalize(self,
                total_counts=False,
                log1p=False,
                scale=False,
                target_sum=None,
                exclude_highly_expressed=False,
                max_fraction=0.05,
                max_value=None,
                inplace=True):
    """
    If ``exclude_highly_expressed=True``, very highly expressed genes are
    excluded from the computation of the normalization factor (size factor)
    for each cell. This is meaningful as these can strongly influence
    the resulting normalized values for all other genes [1]_.

    Parameters
    ----------
    total_counts : bool (default=False)
      Normalize counts per cell.
    log1p : bool (default=False)
      Logarithmize the data matrix.
    scale : bool (default=False)
      Scale data to unit variance and zero mean.
    target_sum : {float, None} (default=None)
      If None, after normalization, each observation (cell) has a
      total count equal to the median of total counts for
      observations (cells) before normalization.
    exclude_highly_expressed : bool (default=False)
      Exclude (very) highly expressed genes for the computation of the
      normalization factor (size factor) for each cell. A gene is considered
      highly expressed, if it has more than ``max_fraction`` of the total counts
      in at least one cell. The not-excluded genes will sum up to
      ``target_sum``.
    max_fraction : bool (default=0.05)
      If ``exclude_highly_expressed=True``, consider cells as highly expressed
      that have more counts than ``max_fraction`` of the original total counts
      in at least one cell.
    max_value : `float` or `None`, optional (default=`None`)
        Clip (truncate) to this value after scaling. If `None`, do not clip.
    inplace : `bool` (default=True)
      if False, return new `SingleCellOMIC` with the filtered
      cells applied

    References
    ----------
    [1] Weinreb et al. (2016), SPRING: a kinetic interface for visualizing
    high dimensional single-cell expression data, bioRxiv.

    Note
    ----
    Proxy to `scanpy.pp.normalize_total`,  `scanpy.pp.log1p`
    and  `scanpy.pp.scale`
    """
    omics = self if inplace else self.copy()

    if total_counts:
      sc.pp.normalize_total(omics,
                            target_sum=target_sum,
                            exclude_highly_expressed=exclude_highly_expressed,
                            max_fraction=max_fraction,
                            inplace=True)
      # since the total counts is normalized, store the old library size
      omics._total_counts = self._total_counts
      omics._log_counts = self._log_counts
      omics._local_mean = self._local_mean
      omics._local_var = self._local_var
      omics._name += '_countnorm'

    if log1p:
      sc.pp.log1p(omics,
                  chunked=True,
                  chunk_size=_DEFAULT_BATCH_SIZE,
                  copy=False)
      omics._name += '_log1p'

    if scale:
      sc.pp.scale(omics, zero_center=True, max_value=max_value, copy=False)
      omics._name += '_scale'
    return omics

  # ====== statistics ====== #
  @property
  def sparsity(self):
    return sparsity_percentage(self.X)

  @property
  def counts_per_cell(self):
    """ Return total number of counts per cell. This method
    is scalable. """
    counts = 0
    for s, e in batching(batch_size=_DEFAULT_BATCH_SIZE, n=self.X.shape[1]):
      counts += np.sum(self.X[:, s:e], axis=1)
    return counts

  @property
  def counts_per_gene(self):
    """ Return total number of counts per gene. This method
    is scalable. """
    counts = 0
    for s, e in batching(batch_size=_DEFAULT_BATCH_SIZE, n=self.X.shape[0]):
      counts += np.sum(self.X[s:e], axis=0)
    return counts

  # ******************** metrics ******************** #
  def neighbors(self,
                n_neighbors=15,
                n_pcs=None,
                use_rep=None,
                knn=True,
                method='umap',
                metric='euclidean',
                seed=8):
    """\
    Compute a neighborhood graph of observations [McInnes18]_.

    The neighbor search efficiency of this heavily relies on UMAP [McInnes18]_,
    which also provides a method for estimating connectivities of data points -
    the connectivity of the manifold (`method=='umap'`). If `method=='gauss'`,
    connectivities are computed according to [Coifman05]_, in the adaption of
    [Haghverdi16]_.

    Parameters
    ----------
    n_neighbors : `int` (default=15)
        The size of local neighborhood (in terms of number of neighboring data
        points) used for manifold approximation. Larger values result in more
        global views of the manifold, while smaller values result in more local
        data being preserved. In general values should be in the range 2 to 100.
        If `knn` is `True`, number of nearest neighbors to be searched. If `knn`
        is `False`, a Gaussian kernel width is set to the distance of the
        `n_neighbors` neighbor.
    n_pcs : {`int`, `None`} (default=None)
        Use this many PCs. If n_pcs==0 use .X if use_rep is None.
    use_rep : {`None`, ‘X’} or any key for .obsm, optional (default=None)
        Use the indicated representation. If None, the representation is
        chosen automatically: for .n_vars < 50, .X is used, otherwise
        ‘X_pca’ is used. If ‘X_pca’ is not present, it’s computed with
        default parameters.
    knn : `bool` (default=True)
        If `True`, use a hard threshold to restrict the number of neighbors to
        `n_neighbors`, that is, consider a knn graph. Otherwise, use a Gaussian
        Kernel to assign low weights to neighbors more distant than the
        `n_neighbors` nearest neighbor.
    seed : `int` (default=8)
        A numpy random seed.
    method : {{'umap', 'gauss', `None`}}  (default: `'umap'`)
        Use 'umap' [McInnes18]_ or 'gauss' (Gauss kernel following [Coifman05]_
        with adaptive width [Haghverdi16]_) for computing connectivities.
    metric : {`str`, `callable`} (default='euclidean')
        A known metric’s name or a callable that returns a distance.

    Returns
    -------
    returns `SingleCellOMIC` with the following:

    **connectivities** : sparse matrix (`.uns['neighbors']`, dtype `float32`)
        Weighted adjacency matrix of the neighborhood graph of data
        points. Weights should be interpreted as connectivities.
    **distances** : sparse matrix (`.uns['neighbors']`, dtype `float32`)
        Instead of decaying weights, this stores distances for each pair of
        neighbors.
    """
    sc.pp.neighbors(self,
                    n_neighbors=n_neighbors,
                    knn=knn,
                    method=method,
                    metric=metric,
                    n_pcs=n_pcs,
                    use_rep=use_rep,
                    random_state=seed,
                    copy=False)
    return self

  def louvain(self,
              resolution=None,
              restrict_to=None,
              key_added='louvain',
              adjacency=None,
              flavor='vtraag',
              directed=True,
              use_weights=False,
              partition_type=None,
              partition_kwargs=None,
              seed=8):
    """Cluster cells into subgroups [Blondel08]_ [Levine15]_ [Traag17]_.

    Cluster cells using the Louvain algorithm [Blondel08]_ in the implementation
    of [Traag17]_. The Louvain algorithm has been proposed for single-cell
    analysis by [Levine15]_.

    This requires having ran :func:`~scanpy.pp.neighbors` or :func:`~scanpy.external.pp.bbknn` first,
    or explicitly passing a ``adjacency`` matrix.

    Parameters
    ----------
    resolution
        For the default flavor (``'vtraag'``), you can provide a resolution
        (higher resolution means finding more and smaller clusters),
        which defaults to 1.0. See “Time as a resolution parameter” in [Lambiotte09]_.
    restrict_to
        Restrict the clustering to the categories within the key for sample
        annotation, tuple needs to contain ``(obs_key, list_of_categories)``.
    key_added
        Key under which to add the cluster labels. (default: ``'louvain'``)
    adjacency
        Sparse adjacency matrix of the graph, defaults to
        ``adata.uns['neighbors']['connectivities']``.
    flavor : {``'vtraag'``, ``'igraph'``}
        Choose between to packages for computing the clustering.
        ``'vtraag'`` is much more powerful, and the default.
    directed
        Interpret the ``adjacency`` matrix as directed graph?
    use_weights
        Use weights from knn graph.
    partition_type
        Type of partition to use.
        Only a valid argument if ``flavor`` is ``'vtraag'``.
    partition_kwargs
        Key word arguments to pass to partitioning,
        if ``vtraag`` method is being used.
    seed
        Change the initialization of the optimization.

    Returns
    -------
    :obj:`None`
        By default (``copy=False``), updates ``adata`` with the following fields:

        ``adata.obs['louvain']`` (:class:`pandas.Series`, dtype ``category``)
            Array of dim (number of samples) that stores the subgroup id
            (``'0'``, ``'1'``, ...) for each cell.

    :class:`~anndata.AnnData`
        When ``copy=True`` is set, a copy of ``adata`` with those fields is returned.
    """
    sc.tl.louvain(self,
                  resolution=resolution,
                  random_state=seed,
                  restrict_to=restrict_to,
                  key_added=key_added,
                  adjacency=adjacency,
                  flavor=flavor,
                  directed=directed,
                  use_weights=use_weights,
                  partition_type=partition_type,
                  partition_kwargs=partition_kwargs,
                  copy=False)
    return self

  def calculate_qc_metrics(self,
                           expr_type='counts',
                           var_type='genes',
                           qc_vars=(),
                           percent_top=(50, 100, 200, 500),
                           layer=None,
                           use_raw=False,
                           parallel=None):
    """\
    Calculate quality control metrics.

    Calculates a number of qc metrics for an AnnData object, see section
    `Returns` for specifics. Largely based on `calculateQCMetrics` from scater
    [McCarthy17]_. Currently is most efficient on a sparse CSR or dense matrix.

    Parameters
    ----------
    expr_type : str
      Name of kind of values in X.
    var_type : str
      The kind of thing the variables are.
    qc_vars : Collection[str]
      Keys for boolean columns of .var which identify variables you could
      want to control for (e.g. “ERCC” or “mito”).
    percent_top : Collection[int]
      Which proportions of top genes to cover. If empty or None don’t
      calculate. Values are considered 1-indexed, percent_top=[50] finds
      cumulative proportion to the 50th most expressed gene.
    layer : str, None
      If provided, use adata.layers[layer] for expression values instead of
      adata.X.
    use_raw : bool
      If True, use adata.raw.X for expression values instead of adata.X.
    parallel : bool, None
      Whether to force parallelism. Otherwise usage of paralleism is based
      on compilation time and sample size heuristics.

    Returns
    -------
    updates `SingleCellOMIC`'s `obs` and `var`.

    Observation level metrics include:
    total_{var_type}_by_{expr_type}
      E.g. “total_genes_by_counts”. Number of genes with positive counts in
      a cell.
    total_{expr_type}
      E.g. “total_counts”. Total number of counts for a cell.
    pct_{expr_type}_in_top_{n}_{var_type} - for n in percent_top
      E.g. “pct_counts_in_top_50_genes”. Cumulative percentage of counts for
      50 most expressed genes in a cell.
    total_{expr_type}_{qc_var} - for qc_var in qc_vars
      E.g. “total_counts_mito”. Total number of counts for variabes in qc_vars.
    pct_{expr_type}_{qc_var} - for qc_var in qc_vars
      E.g. “pct_counts_mito”. Proportion of total counts for a cell which are
      mitochondrial.

    Variable level metrics include:
    total_{expr_type}
      E.g. “total_counts”. Sum of counts for a gene.
    mean_{expr_type}
      E.g. “mean counts”. Mean expression over all cells.
    n_cells_by_{expr_type}
      E.g. “n_cells_by_counts”. Number of cells this expression is measured in.
    pct_dropout_by_{expr_type}
      E.g. “pct_dropout_by_counts”. Percentage of cells this feature does not
      appear in.
    """
    sc.pp.calculate_qc_metrics(self,
                               expr_type=expr_type,
                               var_type=var_type,
                               qc_vars=qc_vars,
                               percent_top=percent_top,
                               layer=layer,
                               use_raw=use_raw,
                               parallel=parallel,
                               inplace=True)
    return self

  # ******************** plotting helper ******************** #
  def plot_percentile_histogram(self,
                                n_hist,
                                title=None,
                                outlier=0.001,
                                non_zeros=False,
                                fig=None):
    """ Data is chopped into multiple percentile (`n_hist`) and the
    histogram is plotted for each percentile.

    """
    from matplotlib import pyplot as plt
    arr = self.X
    if non_zeros:
      arr = arr[arr != 0]
    n_percentiles = n_hist + 1
    n_col = 5
    n_row = int(np.ceil(n_hist / n_col))
    if fig is None:
      fig = vs.plot_figure(nrow=int(n_row * 1.5), ncol=20)
    self.assert_figure(fig)
    percentile = np.linspace(start=np.min(arr),
                             stop=np.max(arr),
                             num=n_percentiles)
    n_samples = len(arr)
    for i, (p_min, p_max) in enumerate(zip(percentile, percentile[1:])):
      min_mask = arr >= p_min
      max_mask = arr <= p_max
      mask = np.logical_and(min_mask, max_mask)
      a = arr[mask]
      _, bins = vs.plot_histogram(
          a,
          bins=120,
          ax=(n_row, n_col, i + 1),
          fontsize=8,
          color='red' if len(a) / n_samples < outlier else 'blue',
          title=("[%s]" % title if i == 0 else "") +
          "%d(samples)  Range:[%g, %g]" % (len(a), p_min, p_max))
      plt.gca().set_xticks(np.linspace(np.min(bins), np.max(bins), num=8))
    plt.tight_layout()
    self.add_figure('percentile_%dhistogram' % n_hist, fig)
    return self

  def plot_heatmap(self, ax=None):
    raise NotImplementedError

  def plot_rank_features_stacked_violin(self, n_features=10, ax=None):
    raise NotImplementedError

  # ******************** logging and io ******************** #
  def save_to_mmaparray(self, path, dtype=None):
    """ This only save the data without row names and column names """
    with MmapArrayWriter(path=path,
                         shape=self.shape,
                         dtype=self.dtype if dtype is None else dtype,
                         remove_exist=True) as f:
      for s, e in batching(batch_size=_DEFAULT_BATCH_SIZE, n=self.n_obs):
        x = self.X[s:e]
        if dtype is not None:
          x = x.astype(dtype)
        f.write(x)

  def _get_str(self):
    all_nonzeros = []
    for s, e in batching(n=self.n_obs, batch_size=_DEFAULT_BATCH_SIZE):
      x = self.X[s:e]
      ids = np.nonzero(x)
      all_nonzeros.append(x[ids[0], ids[1]])
    all_nonzeros = np.concatenate(all_nonzeros)

    text = super(SingleCellOMIC, self).__repr__()
    text = text.replace('AnnData object', self.name)
    text += '\n    Sparsity: %.2f' % self.sparsity
    text += '\n    Nonzeros: %s' % describe(
        all_nonzeros, shorten=True, float_precision=2)
    text += '\n    Cell    : %s' % describe(
        self.counts_per_cell, shorten=True, float_precision=2)
    text += '\n    Gene    : %s' % describe(
        self.counts_per_gene, shorten=True, float_precision=2)
    text += '\n    TotalCount: %s' % describe(
        self.total_counts, shorten=True, float_precision=2)
    text += '\n    LogCount  : %s' % describe(
        self.log_counts, shorten=True, float_precision=2)
    text += '\n    LocalMean : %s' % describe(
        self.local_mean, shorten=True, float_precision=2)
    text += '\n    LocalVar  : %s' % describe(
        self.local_var, shorten=True, float_precision=2)
    return text

  def __repr__(self):
    return self._get_str()

  def __str__(self):
    return self._get_str()



from __future__ import absolute_import, division, print_function

import tensorflow as tf
from odin.networks import DenseDeterministic
from sisua.models.base import SingleCellModel
from sisua.models.networks import DenseNetwork

class SCScope(SingleCellModel):

  def __init__(self,
               units,
               steps=1,
               dispersion='full',
               xdist='zinbd',
               zdist='normal',
               ldist='normal',
               xdrop=0.3,
               edrop=0,
               zdrop=0,
               ddrop=0,
               hdim=128,
               zdim=32,
               nlayers=2,
               clip_library=12,
               batchnorm=True,
               linear_decoder=False,
               **kwargs):
    super(SCScope, self).__init__(units=units,
                                  xdist=xdist,
                                  dispersion=dispersion,
                                  parameters=locals(),
                                  **kwargs)
    self.steps = int(steps)
    self.encoder = DenseNetwork(n_units=hdim,
                                nlayers=nlayers,
                                activation='relu',
                                batchnorm=batchnorm,
                                input_dropout=xdrop,
                                output_dropout=edrop,
                                seed=self.seed,
                                name='Encoder')
    if linear_decoder:
      self.decoder = Identity(name="Decoder")
    else:
      self.decoder = DenseNetwork(n_units=hdim,
                                  nlayers=nlayers,
                                  activation='relu',
                                  batchnorm=batchnorm,
                                  input_dropout=zdrop,
                                  output_dropout=ddrop,
                                  seed=self.seed,
                                  name='Decoder')
    self.latent_layer = DenseDeterministic(zdim,
                                           use_bias=bool(biased_latent),
                                           activation='linear',
                                           name='Latent')

  def _call(self, x, lmean, lvar, t, y, masks, training=None, n_samples=None):
    exit()



from __future__ import absolute_import, division, print_function

import os
import pickle
import shutil

import numpy as np

from odin.fuel import Dataset
from odin.utils import ctext, one_hot, select_path
from sisua.data.path import DOWNLOAD_DIR, PREPROCESSED_BASE_DIR



def _save_data_to_path(preprocessed_path, X, y, gene_names, label_names,
                       cell_names, verbose):
  # save data
  if verbose:
    print("Saving data to %s ..." % ctext(preprocessed_path, 'cyan'))
  with open(os.path.join(preprocessed_path, 'X'), 'wb') as f:
    pickle.dump(X, f)
  with open(os.path.join(preprocessed_path, 'y'), 'wb') as f:
    pickle.dump(y, f)
  # save the meta info
  with open(os.path.join(preprocessed_path, 'X_row'), 'wb') as f:
    pickle.dump(cell_names, f)
  with open(os.path.join(preprocessed_path, 'X_col'), 'wb') as f:
    pickle.dump(gene_names, f)
  with open(os.path.join(preprocessed_path, 'y_col'), 'wb') as f:
    pickle.dump(label_names, f)


# ===========================================================================
# Common dataset
# ===========================================================================
def _read_scvi_dataset(name, clazz_name, override, verbose):
  preprocessed_path = select_path(os.path.join(PREPROCESSED_BASE_DIR,
                                               '%s_preprocessed' % name),
                                  create_new=True)
  if override:
    shutil.rmtree(preprocessed_path)
    os.mkdir(preprocessed_path)
  # ====== copy the dataset from scVI ====== #
  if not os.path.exists(os.path.join(preprocessed_path, 'X')):
    try:
      import scvi.dataset as scvi_dataset
    except ImportError:
      raise RuntimeError("Require `scVI` package for PBMC dataset")
    clazz = getattr(scvi_dataset, clazz_name)
    gene_dataset = clazz(save_path=DOWNLOAD_DIR)

    X = gene_dataset._X
    if hasattr(X, 'todense'):
      X = np.array(X.todense())

    gene_names = np.array(gene_dataset.gene_names)
    # convert gene identifier to gene symbol (i.e. name)
    if hasattr(gene_dataset, 'de_metadata'):
      from sisua.data.utils import get_gene_id2name
      meta = gene_dataset.de_metadata
      converter = {i: j for i, j in zip(meta.ENSG, meta.GS)}
      pbmc8kconverter = get_gene_id2name()
      gene_names = np.array([
          pbmc8kconverter[i] if i in pbmc8kconverter else converter[i]
          for i in gene_names
      ])
    assert len(gene_names) == X.shape[1]

    label_names = np.array(gene_dataset.cell_types)
    y = one_hot(gene_dataset.labels.ravel(), nb_classes=len(label_names))
    assert len(label_names) == y.shape[1]

    cell_names = np.array(['Cell#%d' % i for i in range(X.shape[0])])
    _save_data_to_path(preprocessed_path, X, y, gene_names, label_names,
                       cell_names, verbose)
  # ====== read preprocessed data ====== #
  ds = Dataset(preprocessed_path, read_only=True)
  return ds


def read_Cortex(override=False, verbose=False):
  return _read_scvi_dataset(name="CORTEX",
                            clazz_name='CortexDataset',
                            override=override,
                            verbose=verbose)


def read_PBMC(override=False, verbose=False):
  return _read_scvi_dataset(name="PBMC_scVI",
                            clazz_name='PbmcDataset',
                            override=override,
                            verbose=verbose)


def read_Retina(override=False, verbose=False):
  return _read_scvi_dataset(name="RETINA",
                            clazz_name='RetinaDataset',
                            override=override,
                            verbose=verbose)


# ===========================================================================
# HEMATO dataset
# ===========================================================================
def read_Hemato(override=False, verbose=False):
  preprocessed_path = select_path(os.path.join(PREPROCESSED_BASE_DIR,
                                               'HEMATO_preprocessed'),
                                  create_new=True)

  if override:
    shutil.rmtree(preprocessed_path)
    os.mkdir(preprocessed_path)
  # ====== copy the dataset from scVI ====== #
  if not os.path.exists(os.path.join(preprocessed_path, 'X')):
    try:
      from scvi.dataset import HematoDataset
    except ImportError:
      raise RuntimeError("Require `scVI` package for HEMATO dataset")

    gene_dataset = HematoDataset(
        save_path=os.path.join(DOWNLOAD_DIR, 'HEMATO/'))

    X = gene_dataset._X
    gene_names = np.array(gene_dataset.gene_names)
    assert len(gene_names) == X.shape[1]

    y = gene_dataset.meta.values[:, 1:]
    label_names = np.array(gene_dataset.cell_types_levels)
    assert len(label_names) == y.shape[1]

    cell_names = np.array(['Cell#%d' % i for i in range(X.shape[0])])

    _save_data_to_path(preprocessed_path, X, y, gene_names, label_names,
                       cell_names, verbose)

    # create a binary classes for testing
    label_names = np.array(["Erythroblasts", "Granulocytes"])
    min_y = np.min(gene_dataset.labels)
    max_y = np.max(gene_dataset.labels)
    y_val = 2 * (gene_dataset.labels - min_y) / (max_y - min_y) - 1
    y_bin = np.argmax(
        np.hstack((
            gene_dataset.meta.iloc[:, 1].values[:, None],  # Er
            gene_dataset.meta.iloc[:, 2].values[:, None])),  # Gr
        axis=-1)
    with open(os.path.join(preprocessed_path, 'labels_name'), 'wb') as f:
      pickle.dump(label_names, f)
    with open(os.path.join(preprocessed_path, 'labels_bin'), 'wb') as f:
      pickle.dump(y_bin, f)
    with open(os.path.join(preprocessed_path, 'labels_val'), 'wb') as f:
      pickle.dump(y_val, f)
  # ====== read preprocessed data ====== #
  ds = Dataset(preprocessed_path, read_only=True)
  return ds



from __future__ import absolute_import, division, print_function

from sisua.utils.io_utils import save_data, save_data_to_csv, save_data_to_R
from sisua.utils.others import *
from sisua.utils.visualization import (fast_scatter, plot_evaluate_classifier,
                                       plot_evaluate_regressor,
                                       plot_evaluate_regressor,
                                       plot_monitoring_epoch, show_image)
from sisua.utils.others import *



# Baseline System using PCA and T-SNE for dimension reduction
from __future__ import print_function, division, absolute_import
import matplotlib
matplotlib.use('Agg')

import os
os.environ['ODIN'] = 'gpu,float32,seed=5218'

import numpy as np

from odin import visual as V
from odin.ml import GMM, PLDA, PPCA, SupervisedPPCA
from odin.visual import (print_dist, merge_text_graph, plot_confusion_matrix,
                         plot_figure, plot_histogram_layers, plot_save,
                         generate_random_colors)
from odin.utils import (unique_labels, ctext, auto_logging, batching, UnitTimer,
                        ArgController, get_script_path, mpi)
from odin.stats import train_valid_test_split

from sklearn.decomposition import PCA
from sklearn.neural_network import MLPRegressor
from sklearn.linear_model import ElasticNetCV
from sklearn.svm import SVC, SVR
from sklearn.manifold import Isomap, SpectralEmbedding, MDS
from sklearn.multiclass import OneVsRestClassifier, OneVsOneClassifier
from sklearn.multioutput import MultiOutputRegressor
from sklearn.mixture import GaussianMixture, BayesianGaussianMixture
from sklearn.cluster import FeatureAgglomeration
from sklearn.metrics import (r2_score, mean_squared_error,
                             mean_absolute_error, explained_variance_score)

from sisua.data import read_CITEseq_PBMC, EXP_DIR
from sisua.data.const import PBMC_markers_to_symbols, PBMC_colors
from sisua.analysis.latent_benchmarks import plot_latents
from sisua.utils import (plot_figure, plot_evaluate_classifier,
                         plot_evaluate_reconstruction, plot_evaluate_regressor)

# ===========================================================================
# Constants
# ===========================================================================
TRAIN_PERCENTAGE = 0.6
NUM_COMPONENTS = 64
# ====== input feature type ====== #
# ['data', 'data_norm', 'data_scale']
DATA = 'data_norm'
# ====== others ====== #
PROTEIN_ORDER = list(PBMC_markers_to_symbols.keys())
mRNA_ORDER = list(PBMC_markers_to_symbols.values())
# ===========================================================================
# Loading data
# ===========================================================================
ds = read_CITEseq_PBMC(override=False)
data = ds[DATA]
gene_name = np.array(ds['cols'][:])
# ====== filter mouse genes ====== #
human_cols = [True if "HUMAN_" in i else False for i in gene_name]
data = data[:, human_cols]
gene_name = np.array([i for i in gene_name if "HUMAN_" in i])
# ====== mRNA ====== #
y_names = ds['meta_cols'][2:12]
ids = {j.replace('mRNA.HUMAN_', ''): i
       for i, j in enumerate(y_names)}
y_mRNA_names = [y_names[ids[i]].replace("mRNA.HUMAN_", "")
                for i in mRNA_ORDER]
y_mRNA = ds['metadata'][:, 2:12][:, [ids[i] for i in mRNA_ORDER]]
# ====== protein.count ====== #
y_prot_names = [i.replace("protein.count.", "")
                for i in ds['meta_cols'][12:22]]
y_prot = ds['metadata'][:, 12:22]
# ====== log ====== #
assert y_mRNA_names == mRNA_ORDER and y_prot_names == PROTEIN_ORDER
# ====== load thresholded protein values ====== #
y_bin = ds['y_bin'] if 'y_bin' in ds else None
y_prob = ds['y_prob'] if 'y_prob' in ds else None
# ===========================================================================
# split the data
# ===========================================================================
num_samples = data.shape[0]
ids = get_rng().permutation(num_samples)
train, valid, test = train_valid_test_split(ids, train=TRAIN_PERCENTAGE, inc_test=True,
                                            seed = get_rng().randint(0, 10e8))
X_train, y_mRNA_train, y_prot_train, y_bin_train, y_prob_train = data[train], y_mRNA[train], y_prot[train], y_bin[train], y_prob[train]
X_valid, y_mRNA_valid, y_prot_valid, y_bin_valid, y_prob_valid = data[valid], y_mRNA[valid], y_prot[valid], y_bin[valid], y_prob[valid]
X_test, y_mRNA_test, y_prot_test, y_bin_test, y_prob_test = data[test], y_mRNA[test], y_prot[test], y_bin[test], y_prob[test]
print(ctext("Train:", 'cyan'), X_train.shape, y_mRNA_train.shape, y_prot_train.shape, y_bin_train.shape, y_prob_train.shape)
print(ctext("Valid:", 'cyan'), X_valid.shape, y_mRNA_valid.shape, y_prot_valid.shape, y_bin_valid.shape, y_prob_valid.shape)
print(ctext("Test:", 'cyan'), X_test.shape, y_mRNA_test.shape, y_prot_test.shape, y_bin_test.shape, y_prob_test.shape)
# ===========================================================================
# Evaluation
# ===========================================================================
random_state = get_rng().randint(0, 10e8)
# ====== create transformer ====== #
with UnitTimer(name="Fit PCA"):
  pca = PCA(n_components=NUM_COMPONENTS, random_state=random_state)
  pca.fit(X_train)

with UnitTimer(name="Fit PPCA"):
  ppca = PPCA(n_components=NUM_COMPONENTS, verbose=True, random_state=random_state)
  ppca.fit(X_train)

with UnitTimer(name="Fit S-PPCA"):
  sppca = SupervisedPPCA(n_components=NUM_COMPONENTS, verbose=True, extractor='supervised',
                         random_state=random_state)
  X_, y_ = [], []
  for i, j in zip(X_train, y_bin_train):
    for _ in range(int(np.sum(j))):
      X_.append(i)
    for c in np.nonzero(j)[0]:
      y_.append(c)
  X_, y_ = np.array(X_), np.array(y_)
  ids = get_rng().permutation(X_.shape[0])
  sppca.fit(X_[ids], y_[ids])

# with UnitTimer(name="Fit GMM"):
#   pca_ = PCA(n_components=800)
#   pca_.fit(X_train)
#   gmm = GaussianMixture(n_components=NUM_COMPONENTS, covariance_type='full',
#                         max_iter=250, init_params='random', n_init=2,
#                         verbose=8, verbose_interval=1,
#                         random_state=random_state)
#   gmm.fit(pca_.transform(X_train))
#   gmm.transform = lambda x: gmm._estimate_log_prob(pca_.transform(x))
# ====== evaluation ====== #
for transformer_name, transformer in [
    ('PCA', pca),
    # ('GMM', gmm),
    ("PPCA", ppca),
    ("SPPCA", sppca)
]:
  print(ctext('Methods:', 'yellow'), ctext(transformer_name, 'cyan'))
  Z_train = transformer.transform(X_train)
  Z_valid = transformer.transform(X_valid)
  Z_test = transformer.transform(X_test)
  # ====== plot T-SNE cell cluster ====== #
  plot_latents_binary(X=Z_test, y=y_prot_test,
               labels=y_prot_names, title='[%s]Test-Prot' % transformer_name,
               ax=224)
  # ====== Training SVM and logistic regression ====== #
  # Classifier
  for name, classifier in [
      ('SVCrbf', OneVsRestClassifier(SVC(kernel='rbf', random_state=random_state))),
  ]:
    print("Testing Classifier:", ctext(name, 'cyan'))
    classifier.fit(X=np.concatenate([Z_train, Z_valid], axis=0),
                   y=np.concatenate([y_bin_train, y_bin_valid], axis=0))
    y_bin_test_pred = classifier.predict(Z_test)
    plot_evaluate_classifier(y_pred=y_bin_test_pred, y_true=y_bin_test,
                             labels=y_prot_names,
                             title='[%s]%s-%s' % (transformer_name, 'Classification', name))
  # Regressor
  for name, classifier in [
      # ('SVRrbf', MultiOutputRegressor(SVR(kernel='rbf'))),
      ('Elastic', MultiOutputRegressor(ElasticNetCV(random_state=random_state))),
  ]:
    print("Testing Regressor:", ctext(name, 'cyan'))
    classifier.fit(X=np.concatenate([Z_train, Z_valid], axis=0),
                   y=np.concatenate([y_prot_train, y_prot_valid], axis=0))
    y_prot_test_pred = classifier.predict(Z_test)
    plot_evaluate_regressor(y_pred=y_prot_test_pred, y_true=y_prot_test,
                            labels=y_prot_names,
                            title='[%s]%s-%s' % (transformer_name, 'Regression', name))
  # Reconstruction
  # for name, classifier in [
  #     ('MLP', MLPRegressor(hidden_layer_sizes=(512, 512, 512), verbose=True,
  #                          random_state=random_state)),
  # ]:
  #   print("Testing Reconstruction:", ctext(name, 'cyan'))
  #   classifier.fit(X=np.concatenate([Z_train, Z_valid], axis=0),
  #                  y=np.concatenate([X_train, X_valid], axis=0))
  #   X_test_pred = classifier.predict(Z_test)
  #   plot_evaluate_reconstruction(X=X_test, X_res=X_test_pred,
  #                                gene_name=gene_name,
  #                                title='[%s]%s-%s' % (transformer_name, 'Reconstruction', name))
  # ====== save all figure ====== #
  V.plot_save(os.path.join(EXP_DIR, 'baseline_%s.png' % transformer_name.lower()))



from __future__ import absolute_import, division, print_function

from typing import Type

import tensorflow as tf
from tensorflow_probability.python.distributions import (LogNormal,
                                                         MultivariateNormalDiag,
                                                         Normal)

from odin.bay.distribution_layers import (LogNormalLayer,
                                          MultivariateNormalLayer, NormalLayer)
from odin.networks import DenseDistribution


class NormalDiagLatent(DenseDistribution):

  def __init__(self, units, use_bias=True, name="LatentSpace"):
    super(NormalDiagLatent, self).__init__(
        units=units,
        posterior=MultivariateNormalLayer(event_size=units,
                                          covariance_type='diag',
                                          softplus_scale=True),
        prior=MultivariateNormalDiag(loc=tf.zeros(shape=units),
                                     scale_identity_multiplier=1),
        use_bias=use_bias,
        name=name)


class NormalLatent(DenseDistribution):

  def __init__(self, units, use_bias=True, name="LatentSpace"):
    super(NormalLatent,
          self).__init__(units=units,
                         posterior=NormalLayer(event_shape=units,
                                               softplus_scale=True),
                         prior=Normal(loc=tf.zeros(shape=units),
                                      scale=tf.ones(shape=units)),
                         use_bias=use_bias,
                         name=name)


class LogNormalLatent(DenseDistribution):

  def __init__(self, units, use_bias=True, name="LatentSpace"):
    super(LogNormalLatent,
          self).__init__(units=units,
                         posterior=LogNormalLayer(event_shape=units,
                                                  softplus_scale=True),
                         prior=LogNormal(loc=tf.zeros(shape=units),
                                         scale=tf.ones(shape=units)),
                         use_bias=use_bias,
                         name=name)


class MixedNormalLatent(DenseDistribution):
  pass


class DirichletLatent(DenseDistribution):
  pass


# ===========================================================================
# Helpers
# ===========================================================================
_latent_map = {
    'normal': NormalLatent,
    'diag': NormalDiagLatent,
    'mixed': MixedNormalLatent,
    'diri': DirichletLatent,
    'lognormal': LogNormalLatent
}


def get_latent(name):
  name = str(name).lower()
  if name not in _latent_map:
    raise ValueError("Only support following latent: %s; but given: '%s'" %
                     (', '.join(_latent_map.keys()), name))
  return _latent_map[name]



from __future__ import absolute_import, division, print_function

import inspect
import os
import pickle
import shutil
import time
from collections import OrderedDict, defaultdict
from itertools import product
from typing import Optional

import numpy as np
import pandas as pd
import seaborn as sns
import tensorflow as tf
from matplotlib import pyplot as plt
from six import string_types

from odin.backend import log_norm
from odin.fuel import Dataset
from odin.ml import fast_pca, fast_tsne
from odin.utils import (as_tuple, cache_memory, catch_warnings_ignore, ctext,
                        flatten_list, md5_checksum)
from odin.utils.mpi import MPI
from odin.visual import (Visualizer, plot_aspect, plot_confusion_matrix,
                         plot_figure, plot_frame, plot_save, plot_scatter,
                         plot_scatter_heatmap, to_axis2D)
from sisua.analysis.imputation_benchmarks import (correlation_scores,
                                                  imputation_mean_score,
                                                  imputation_score,
                                                  imputation_std_score)
from sisua.analysis.latent_benchmarks import (clustering_scores,
                                              plot_distance_heatmap,
                                              plot_latents_binary,
                                              plot_latents_multiclasses,
                                              streamline_classifier)
from sisua.analysis.sc_metrics import _preprocess_output_distribution
from sisua.data import SingleCellOMIC, apply_artificial_corruption, get_dataset
from sisua.data.path import EXP_DIR
from sisua.data.utils import standardize_protein_name
from sisua.models.base import SingleCellModel
from sisua.models.scvi_models import SCVI
from sisua.utils import filtering_experiment_path


# ===========================================================================
# The Posterior
# ===========================================================================
class Posterior(Visualizer):
  """ Posterior

  Parameters
  ----------
  scm : {`sisua.models.SingleCellModel`, string}

  n_samples : int
    number of MCMC samples for evaluation

  verbose : bool
    turn on verbose

  Example
  -------
  >>> pos1 = Posterior(mae, x_test, y_test)
  >>> print(pos1.scores_classifier(x_train, y_train))
  >>> print(pos1.scores_llk())
  >>> print(pos1.scores_imputation())
  >>> print(pos1.scores_spearman())
  >>> print(pos1.scores_pearson())
  >>> print(pos1.scores_clustering())
  >>> pos1.save_scores('/tmp/tmp1.txt')
  ...
  >>> pos1.plot_protein_scatter().plot_protein_predicted_series()
  >>> pos1.plot_classifier_F1(x_train, y_train)
  >>> pos1.plot_learning_curves('loss_x')
  >>> pos1.plot_learning_curves()
  >>> pos1.get_correlation_marker_pairs('X')
  >>> pos1.get_correlation_marker_pairs('V')
  >>> pos1.get_correlation_marker_pairs('W')
  >>> pos1.get_correlation_marker_pairs('T')
  >>> pos1.plot_correlation_top_pairs().plot_correlation_bottom_pairs()
  >>> pos1.plot_latents_binary_scatter().plot_latents_distance_heatmap(
  >>> ).plot_latents_protein_pairs()
  >>> pos1.save_figures('/tmp/tmp1.pdf')

  """

  def __init__(self,
               scm: SingleCellModel,
               gene: SingleCellOMIC,
               protein: Optional[SingleCellOMIC] = None,
               batch_size=16,
               n_samples=10,
               verbose=True):
    super(Posterior, self).__init__()
    self.verbose = bool(verbose)
    self.n_samples = int(n_samples)
    self.batch_size = int(batch_size)

    if isinstance(scm, string_types):
      with open(scm, 'rb') as f:
        scm = pickle.load(f)
    assert isinstance(scm, SingleCellModel), \
      "scm must be instance of SingleCellModel but given %s" % str(type(scm))
    self.scm = scm
    # gene expression data
    assert isinstance(gene, SingleCellOMIC), \
      "gene must be instance of SingleCellOMIC"
    self.gene = gene
    # protein level data
    assert isinstance(protein, (SingleCellOMIC, type(None))), \
      "protein must be instance of SingleCellOMIC or None"
    if protein is not None:
      if 'pbe' not in protein.uns:
        protein.probabilistic_embedding()
    self.protein = protein
    # corrupted gene expression
    gene_corrupt = self.gene.corrupt(corruption_rate=self.scm.corruption_rate,
                                     corruption_dist=self.scm.corruption_dist,
                                     inplace=False)
    self.gene_corrupt = gene_corrupt
    # initialize all the prediction
    self._initialize()

  def _initialize(self):
    scm = self.scm
    gene = self.gene
    gene_corrupt = self.gene_corrupt
    protein = self.protein

    # NOTE: for now we haven't find the use of input clean yet
    # save some processing time here
    # outputs_clean, latents_clean = scm.predict(gene,
    #                                            apply_corruption=False,
    #                                            n_samples=self.n_samples,
    #                                            batch_size=self.batch_size,
    #                                            enable_cache=False,
    #                                            verbose=self.verbose)
    outputs_clean = []
    latents_clean = []
    if not isinstance(outputs_clean, (tuple, list)):
      outputs_clean = [outputs_clean]
    if not isinstance(latents_clean, (tuple, list)):
      latents_clean = [latents_clean]

    outputs_corrupt, latents_corrupt = scm.predict(gene_corrupt,
                                                   apply_corruption=False,
                                                   n_samples=self.n_samples,
                                                   batch_size=self.batch_size,
                                                   enable_cache=False,
                                                   verbose=self.verbose)
    if not isinstance(outputs_corrupt, (tuple, list)):
      outputs_corrupt = [outputs_corrupt]
    if not isinstance(latents_corrupt, (tuple, list)):
      latents_corrupt = [latents_corrupt]

    self.outputs_clean = outputs_clean
    self.latents_clean = latents_clean
    self.outputs_corrupt = outputs_corrupt
    self.latents_corrupt = latents_corrupt

  # ******************** Basic matrices ******************** #
  @property
  def gene_name(self):
    return self.gene.var['geneid'].values

  @property
  def protein_name(self):
    name = self.protein.var['protid'].values
    return [standardize_protein_name(i) for i in name]

  @property
  def gene_dim(self):
    return self.gene.shape[1]

  @property
  def prot_dim(self):
    return self.protein.shape[1]

  @property
  def X(self):
    """ Original gene expression data without artificial corruption """
    return self.gene.X

  @property
  def T(self):
    """ The artificial corrupted gene expression similar to the one
    used for fitting the model """
    return self.gene_corrupt.X

  @property
  def W(self):
    return tf.reduce_mean(self.outputs_corrupt[0].mean(), axis=0).numpy()

  @property
  def V(self):
    dist = _preprocess_output_distribution(self.outputs_corrupt[0])
    return tf.reduce_mean(dist.mean(), axis=0).numpy()

  @property
  def Z(self):
    """ Latent space of the autoencoder"""
    return self.latents_corrupt[0].mean().numpy()

  @property
  def y_true(self):
    """ True value of protein marker level or cell type labels """
    return self.protein.X

  @property
  def y_prob(self):
    """ Return the probabilized value of `y_true` """
    if self.is_binary_classes:
      return self.y_true
    return self.protein.obsm['X_prob']

  @property
  def y_bin(self):
    """ Return the binarized value of `y_true` """
    if self.is_binary_classes:
      return self.y_true
    return self.protein.obsm['X_bin']

  @property
  def y_pred(self):
    """ Labels prediction from semi-supervised learning """
    if not self.is_semi_supervised:
      raise RuntimeError(
          "Not semi supervised model, cannot generate prediction for protein values"
      )
    return tf.reduce_mean(self.outputs_corrupt[1].mean(), axis=0).numpy()

  @property
  def name(self):
    i = self.scm.id
    ds_name = self.gene.name.split('_')[0]
    n_gene = '%dgene' % self.X.shape[1]
    n_prot = '%dprot' % (self.y_true.shape[1]
                         if self.protein is not None else 0)
    norm = 'log' if self.scm.log_norm else 'raw'
    corruption_dist = self.scm.corruption_dist
    corruption_rate = self.scm.corruption_rate
    if 'bi' in corruption_dist:
      corruption = 'bi' + str(int(corruption_rate * 100))
    else:
      corruption = 'un' + str(int(corruption_rate * 100))
    if self.is_semi_supervised:
      semi = 'semi'
    else:
      semi = 'unsp'
    return '_'.join([ds_name, i, norm, n_gene, n_prot, corruption, semi])

  @property
  def name_lines(self):
    """same as short_id but divided into 3 lines"""
    short_id = self.name.split('_')
    return '\n'.join(
        ['_'.join(short_id[:2]), '_'.join(short_id[2:-1]), short_id[-1]])

  @property
  def is_semi_supervised(self):
    return self.scm.is_semi_supervised

  @property
  def is_binary_classes(self):
    return np.all(np.sum(self.y_true, axis=-1) == 1.)

  # ******************** helpers ******************** #
  def _train_data(self, x_train, y_train):
    """ return outputs, latents, y_train """
    if not isinstance(x_train, SingleCellOMIC):
      x_train = SingleCellOMIC(x_train)
    if not isinstance(y_train, SingleCellOMIC):
      y_train = SingleCellOMIC(y_train)
    # get binary values for y_train
    if not y_train.is_binary:
      if 'pbe' not in y_train.uns:
        y_train.probabilistic_embedding()
      y_train = y_train.obsm['X_bin']
    else:
      y_train = y_train.X
    # create corrupted data and get the outputs and latents
    x_train = x_train.corrupt(corruption_rate=self.scm.corruption_rate,
                              corruption_dist=self.scm.corruption_dist,
                              inplace=False)
    outputs, latents = self.scm.predict(x_train,
                                        n_samples=self.n_samples,
                                        batch_size=self.batch_size,
                                        enable_cache=False,
                                        verbose=self.verbose)
    # return, always a list, except the labels
    if not isinstance(outputs, (tuple, list)):
      outputs = [outputs]
    if not isinstance(latents, (tuple, list)):
      latents = [latents]
    return outputs, latents, y_train

  # ******************** Latent space analysis ******************** #
  def plot_latents_protein_pairs(self, legend=True, pca=False):
    """ Using marker gene/protein to select mutual exclusive protein
    pairs for comparison """
    z, y = self.Z, self.y_true
    title = self.name
    fig = plot_latents_multiclasses(Z=z,
                                    y=y,
                                    labels_name=self.protein_name,
                                    title=title,
                                    use_PCA=bool(pca),
                                    show_colorbar=bool(legend))
    if fig is not None:
      self.add_figure('latents_protein_pairs', fig)
    return self

  def plot_latents_distance_heatmap(self, legend=True, ax=None, fig=(8, 8)):
    """ Heatmap of the distance among latents vector from different classes
    """
    ax = to_axis2D(ax, fig)
    z, y = self.Z, self.y_true
    title = self.name
    if not self.is_binary_classes:
      from sisua.label_threshold import ProbabilisticEmbedding
      y = ProbabilisticEmbedding().fit_transform(y, return_probabilities=True)
    plot_distance_heatmap(z,
                          labels=y,
                          labels_name=self.protein_name,
                          legend_enable=bool(legend),
                          ax=ax,
                          fontsize=8,
                          legend_ncol=2,
                          title=title)
    self.add_figure('latents_distance_heatmap', ax.get_figure())
    return self

  def plot_latents_binary_scatter(self,
                                  legend=True,
                                  pca=False,
                                  size=8,
                                  ax=None,
                                  fig=(8, 8)):
    """
    test : if True, plotting latent space of test set, otherwise, use training set
    """
    ax = to_axis2D(ax, fig)
    z, y = self.Z, self.y_true
    title = self.name_lines
    plot_latents_binary(Z=z,
                        y=y,
                        title=title,
                        show_legend=bool(legend),
                        size=8,
                        fontsize=8,
                        ax=ax,
                        labels_name=self.protein_name,
                        use_PCA=pca,
                        enable_argmax=True,
                        enable_separated=False)
    self.add_figure('latents_scatter', ax.get_figure())
    return self

  # ******************** Streamline classifier ******************** #
  def plot_classifier_F1(self,
                         x_train: Optional[SingleCellOMIC] = None,
                         y_train: Optional[SingleCellOMIC] = None,
                         plot_train_results=False,
                         mode='ovr'):
    """
    ovr - one vs rest
    ovo - one vs one
    """
    if mode == 'ovo':
      raise NotImplementedError

    if x_train is not None and y_train is not None:
      _, latents, y_train = self._train_data(x_train, y_train)
      Z_train = latents[0].mean().numpy()
    else:
      Z_train = self.Z
      y_train = self.y_bin
    Z_test, y_test = self.Z, self.y_bin

    (train, test), (fig_train, fig_test) = streamline_classifier(
        Z_train,
        y_train,
        Z_test,
        y_test,
        plot_train_results=plot_train_results,
        labels_name=self.protein_name,
        show_plot=True,
        return_figure=True)
    if plot_train_results:
      self.add_figure('streamline_f1_%s' % 'train', fig_train)
    self.add_figure('streamline_f1_%s' % 'test', fig_test)
    return self

  # ******************** Cellsize analysis ******************** #
  def plot_cellsize_series(self, fontsize=10, ax=None, fig=None):
    ax = to_axis2D(ax, fig)
    mean, std, x = self.L_test, self.Lstddev_test, self.X_test_org
    mean = mean.ravel()
    std = std.ravel()
    cell_size = np.sum(x, axis=-1)
    sorted_ids = np.argsort(cell_size)

    ax.plot(cell_size[sorted_ids], linewidth=1, label="Original")
    ax.plot(mean[sorted_ids],
            linestyle='--',
            alpha=0.66,
            linewidth=1,
            label='Prediction')

    ax.set_title('[%s]%s' % ('Test' if test else 'Train', self.name),
                 fontsize=fontsize)
    ax.set_ylabel('Cell Size')
    ax.set_xlabel('Cell in sorted order of size')
    ax.legend()

    self.add_figure('cellsize_%s' % ('test' if test else 'train'),
                    ax.get_figure())
    return self

  # ******************** Correlation analysis ******************** #
  def plot_correlation_top_pairs(self,
                                 data_type='V',
                                 proteins=None,
                                 n=8,
                                 fontsize=10,
                                 fig=None):
    return self._plot_correlation_ranked_pairs(data_type=data_type,
                                               n=n,
                                               proteins=proteins,
                                               top=True,
                                               fontsize=fontsize,
                                               fig=fig)

  def plot_correlation_bottom_pairs(self,
                                    data_type='V',
                                    proteins=None,
                                    n=8,
                                    fontsize=10,
                                    fig=None):
    return self._plot_correlation_ranked_pairs(data_type=data_type,
                                               n=n,
                                               proteins=proteins,
                                               top=False,
                                               fontsize=fontsize,
                                               fig=fig)

  def _plot_correlation_ranked_pairs(self,
                                     data_type='V',
                                     n=8,
                                     proteins=None,
                                     top=True,
                                     fontsize=10,
                                     fig=None):
    """
    Parameters
    ----------
    data_type : {'X', 'T', 'V', 'W'}
      'X' for input gene expression,
      'T' for corrupted gene expression,
      'V' for imputed gene expression,
      and 'W' for reconstructed gene expression
    proteins : {None, 'marker', list of string}

    """
    ax = to_axis2D(ax=None, fig=fig)
    fig = ax.get_figure()

    correlations = self.get_correlation_all_pairs(data_type=data_type)
    y = self.y_true
    if data_type == 'V':
      ydata = self.V
      data_type_name = "Imputed"
    elif data_type == 'X':
      ydata = self.X
      data_type_name = "Original"
    elif data_type == 'T':
      ydata = self.T
      data_type_name = "Corrupted"
    elif data_type == 'W':
      ydata = self.W
      data_type_name = "Reconstructed"

    n = int(n)
    if isinstance(proteins,
                  string_types) and proteins.lower().strip() == 'marker':
      from sisua.data.const import MARKER_GENES
      proteins = [
          i for i in self.protein_name
          if standardize_protein_name(i) in MARKER_GENES
      ]
    elif proteins is None:
      proteins = self.protein_name
    proteins = as_tuple(proteins, t=string_types)

    labels = {
        standardize_protein_name(j).lower(): i
        for i, j in enumerate(self.protein_name)
    }
    prot_ids = []
    for i in proteins:
      i = standardize_protein_name(i).lower()
      if i in labels:
        prot_ids.append(labels[i])
    prot_ids = set(prot_ids)

    # mapping protein_id -> (gene, pearson, spearman)
    correlations_map = defaultdict(list)
    for gene_id, prot_id, pearson, spearman in correlations:
      if prot_id in prot_ids:
        correlations_map[prot_id].append((gene_id, pearson, spearman))
    correlations_map = {
        i: j[:n] if top else j[-n:][::-1] for i, j in correlations_map.items()
    }

    # ====== create figure ====== #
    nrow = len(correlations_map)
    if nrow == 0:  # no matching protein found
      return self

    ncol = n
    if fig is None:
      fig = plot_figure(nrow=3 * nrow, ncol=4 * ncol)
    for i, (prot_idx, data) in enumerate(correlations_map.items()):
      prot = self.protein_name[prot_idx]
      for j, (gene_idx, pearson, spearman) in enumerate(data):
        ax = plt.subplot(nrow, ncol, i * ncol + j + 1)
        gene = self.gene_name[gene_idx]
        sns.scatterplot(x=y[:, prot_idx], y=ydata[:, gene_idx], ax=ax)

        title = 'Pearson:%.2f Spearman:%.2f' % (pearson, spearman)
        ax.set_title(title, fontsize=fontsize)
        if j == 0:
          ax.set_ylabel('Protein:%s' % prot, fontsize=fontsize + 2)
        ax.set_xlabel('Gene:%s' % gene)

    # ====== store the figure ====== #
    plt.suptitle('[data_type: %s] %s %d pairs' %
                 (data_type_name, 'Top' if top else 'Bottom', n),
                 fontsize=fontsize + 8)
    with catch_warnings_ignore(UserWarning):
      plt.tight_layout(rect=[0, 0.03, 1, 0.96])
    self.add_figure(
        'correlation_%s_%s%d' %
        (data_type_name.lower(), 'top' if top else 'bottom', n), fig)
    return self

  def plot_correlation_marker_pairs(self,
                                    imputed=True,
                                    fontsize=10,
                                    proteins=None,
                                    fig=None):
    """ Plotting the correlation series between marker gene and protein,
    The original (uncorrupted data) is used for comparison to the
    imputation.

    Parameters
    ----------
    imputed : `bool` (default: True)
      if `True`, plot the imputed value (in case of zero-inflated distribution),
      otherwise, plot the reconstructed value
    proteins : {`None`, `str`, list of `str`} (default=`None`)
      a list of protein names to be included, if `None`, include all marker
      proteins.
    """
    from scipy.stats import pearsonr, spearmanr
    if proteins is not None:
      proteins = [
          standardize_protein_name(i).lower() for i in as_tuple(proteins, t=str)
      ]

    if imputed:
      v = self.V
      name = 'Imputed'
    else:
      v = self.W
      name = "Reconstructed"

    x, y = self.X, self.y_true
    original_series = correlation_scores(X=x,
                                         y=y,
                                         gene_name=self.gene_name,
                                         protein_name=self.protein_name,
                                         return_series=True)
    imputed_series = correlation_scores(X=v,
                                        y=y,
                                        gene_name=self.gene_name,
                                        protein_name=self.protein_name,
                                        return_series=True)
    # only select given protein
    if proteins is not None:
      original_series = {
          i: j
          for i, j in original_series.items()
          if i.split('/')[0].lower() in proteins
      }
      imputed_series = {
          i: j
          for i, j in original_series.items()
          if i.split('/')[0].lower() in proteins
      }

    assert len(original_series) == len(imputed_series)
    n_pair = len(imputed_series)
    if n_pair == 0:
      return self

    if fig is None:
      fig = plt.figure(figsize=(15, 5 * n_pair), constrained_layout=True)
    assert isinstance(fig, plt.Figure), \
    "fig must be instance of matplotlib.Figure"

    width = 4
    grids = fig.add_gridspec(n_pair, 2 * width)

    for idx, prot_gene in enumerate(sorted(imputed_series.keys())):
      prot_name, gene_name = prot_gene.split('/')
      imputed_gene, prot1 = imputed_series[prot_gene]
      original_gene, prot2 = original_series[prot_gene]
      assert np.all(prot1 == prot2)
      y = prot1

      for j, (name, series) in enumerate(
          (("Original", original_gene), ("Imputed", imputed_gene))):
        ax = fig.add_subplot(grids[idx, width * j:(width * j + width - 1)])

        # plot the points
        ax.scatter(y, series, s=25, alpha=0.6, linewidths=0)
        plot_aspect('auto', 'box', ax)

        # annotations
        ax.set_title('[%s]Pearson:%.2f Spearman:%.2f' % (
            name,
            pearsonr(series, y)[0],
            spearmanr(series, y).correlation,
        ),
                     fontsize=fontsize)
        ax.set_xlabel('[Protein] %s' % prot_name, fontsize=fontsize)
        ax.set_ylabel('[Gene] %s' % gene_name, fontsize=fontsize)

        # box plot for the distribution
        ax = fig.add_subplot(grids[idx, (width * j + width - 1):(width * j +
                                                                 width)])
        ax.boxplot(series)
        ax.set_xticks(())
        ax.set_xlabel(name, fontsize=fontsize)

    plt.suptitle('[%s]%s' % (name, self.name), fontsize=fontsize + 2)
    with catch_warnings_ignore(UserWarning):
      plt.tight_layout(rect=[0, 0.03, 1, 0.96])
    self.add_figure('correlation_marker_%s' % name.lower(), fig)
    return self

  # ******************** Simple analysis ******************** #
  def scores_llk(self):
    """ Return the 'llk' (log-likelihood) of the output distribution
    and the corrupted data used, also includes the 'llk_imputed' which
    is the log-likelihood of the denoised distribution and the original
    data without corruption.
    """
    llk = tf.reduce_mean(self.outputs_corrupt[0].log_prob(self.T)).numpy()
    llk_imputed = tf.reduce_mean(
        _preprocess_output_distribution(self.outputs_corrupt[0]).log_prob(
            self.X)).numpy()
    return {'llk': llk, 'llk_imputed': llk_imputed}

  def scores_imputation(self):
    """(train_score, train_score_mean, train_score_std),
       (test_score, test_score_mean, test_score_std)"""
    return {
        'all': imputation_score(self.X, self.V),
        'mean': imputation_mean_score(self.X, self.T, self.V),
        'std': imputation_std_score(self.X, self.T, self.V),
    }

  def scores_spearman(self):
    return self.get_correlation_marker_pairs(data_type='V',
                                             score_type='spearman')

  def scores_pearson(self):
    return self.get_correlation_marker_pairs(data_type='V',
                                             score_type='pearson')

  def scores_clustering(self):
    return clustering_scores(latent=self.Z,
                             labels=self.y_true,
                             n_labels=len(self.protein_name))

  def scores_classifier(self,
                        x_train: Optional[SingleCellOMIC] = None,
                        y_train: Optional[SingleCellOMIC] = None):
    if x_train is not None and y_train is not None:
      _, latents, y_train = self._train_data(x_train, y_train)
      Z_train = latents[0].mean().numpy()
    else:
      Z_train = self.Z
      y_train = self.y_bin
    Z_test, y_test = self.Z, self.y_bin
    return streamline_classifier(Z_train,
                                 y_train,
                                 Z_test,
                                 y_test,
                                 train_results=False,
                                 labels_name=self.protein_name,
                                 show_plot=False)

  def save_scores(self, path=None):
    """ Saving all scores to a txt file """
    text = '==== %s ====\n' % self.name
    for name, scores in (
        ('llk', self.scores_llk()),
        ('pearson', self.scores_pearson()),
        ('spearman', self.scores_spearman()),
        ('classifier', self.scores_classifier()),
        ('cluster', self.scores_clustering()),
        ('imputation', self.scores_imputation()),
    ):
      text += '%s\n' % name
      for i, j in sorted(scores.items()):
        j = '%+.4f' % j
        j = j.replace('+', ' ')
        text += ' %12s :%s\n' % (i, j)
    if path is None:
      return text
    with open(path, 'w') as f:
      f.write(text)
    return self

  # ******************** learning curves and metrics ******************** #
  @property
  def learning_curves(self):
    return {
        'train': self.train_history['loss'],
        'valid': self.valid_history['val_loss'],
    }

  @property
  def train_history(self):
    return self.scm.train_history

  @property
  def valid_history(self):
    return self.scm.valid_history

  def plot_learning_curves(self,
                           metrics='loss',
                           ignore_case=True,
                           ignore_missing=True,
                           fig=(8, 8)):
    """ Plotting the loss or metrics returned during training progress

    Parameters
    ----------
    metrics : {`str`, list of `str`} (default='loss')
      a metric name or list of metric name for plotting
    ignore_case : `bool` (default=`False`)
      if `True` turn everything to lower case for matching the metric name.
    ignore_missing : `bool` (default=`True`)
      if `True`, ignore the missing metric which not found in the history of
      trained model.
    """
    metrics = as_tuple(metrics, t=str)
    if ignore_case:
      metrics = [m.lower() for m in metrics]
    ax = to_axis2D(None, fig)
    fig = ax.get_figure()

    train_history = self.train_history
    valid_history = self.valid_history

    _ = []
    train_keys = {
        i.lower() if ignore_case else i: i for i in train_history.keys()
    }
    valid_keys = {
        i.lower() if ignore_case else i: i for i in valid_history.keys()
    }
    for m in metrics:
      if not (m in train_keys and 'val_' + m in valid_keys):
        if not ignore_missing:
          raise ValueError(
              "Cannot find metric with name: %s; all given metrics are: %s" %
              (m, ', '.join(train_keys)))
      else:
        _.append(train_keys[m])

    metrics = tuple(_)
    n_metrics = len(metrics)
    if n_metrics == 0:  # no metric or loss found for plotting
      return self

    for idx, m in enumerate(metrics):
      ax = plt.subplot(1, n_metrics, idx + 1)

      train = train_history[m]
      valid = valid_history['val_' + m]

      line_styles = dict(linewidth=1.8)
      point_styles = dict(alpha=0.6, s=80, linewidths=0)

      ax.plot(train, label='train', color='blue', linestyle='-', **line_styles)
      ax.scatter(np.argmin(train), np.min(train), c='blue', **point_styles)

      ax.plot(valid,
              label='valid',
              color='orange',
              linestyle='--',
              **line_styles)
      ax.scatter(np.argmin(valid), np.min(valid), c='orange', **point_styles)

      # ax.set_ylabel("Loss")
      ax.set_xlabel("#Epoch")
      # convert all the xticks to integer and remove overlapping
      xticks = ax.get_xticks()
      ax.set_xticks(list(set([int(i) for i in xticks])))

      ax.legend()
      ax.set_title('[metric: %s]' % m, fontsize=14)

    fig.suptitle(self.name)
    with catch_warnings_ignore(UserWarning):
      plt.tight_layout(rect=[0, 0.03, 1, 0.96])
    self.add_figure('learning_curves_%s' % m, ax.get_figure())
    return self

  # ******************** Correlation scores ******************** #
  def get_correlation_marker_pairs(self, data_type='X', score_type='spearman'):
    """
    Parameters
    ----------
    data_type : {'X', 'V', 'W', 'T'}
      'X' for input gene expression,
      'T' for corrupted gene expression,
      'V' for imputed gene expression,
      and 'W' for reconstructed gene expression

    score_type : {'spearman', 'pearson'}
      spearman correlation for rank (monotonic) relationship, and pearson
      for linear correlation

    Return
    ------
    correlation : OrderedDict
      mapping from marker protein/gene name (string) to
      correlation score (scalar)
    """
    assert score_type in ('spearman', 'pearson')
    assert data_type in ('X', 'V', 'W', 'T')
    y = self.y_true
    X = getattr(self, data_type)
    corr = correlation_scores(X=X,
                              y=y,
                              gene_name=self.gene_name,
                              protein_name=self.protein_name,
                              return_series=False)
    score_idx = 0 if score_type == 'spearman' else 1
    return OrderedDict([(i, j[score_idx]) for i, j in corr.items()])

  @cache_memory
  def get_correlation_all_pairs(self, data_type='X'):
    """
    Parameters
    ----------
    data_type : {'X', 'T', 'V', 'W'}
      'X' for input gene expression,
      'T' for corrupted gene expression,
      'V' for imputed gene expression,
      and 'W' for reconstructed gene expression

    Return
    ------
    correlation : tuple of four scalars
      list of tuple contained 4 scalars
      (gene-idx, protein-idx, pearson, spearman)
      sorted in order from high to low correlation
    """
    assert data_type in ('X', 'V', 'W', 'T')

    from scipy.stats import pearsonr, spearmanr
    v, x, t, w, y = self.V, self.X, self.T, self.W, self.y_true

    n_proteins = y.shape[1]
    n_genes = x.shape[1]
    data = getattr(self, data_type)

    # ====== search for most correlated series ====== #
    def _corr(idx):
      for gene_idx, prot_idx in idx:
        g = data[:, gene_idx]
        p = y[:, prot_idx]
        with catch_warnings_ignore(RuntimeWarning):
          yield (gene_idx, prot_idx, pearsonr(g,
                                              p)[0], spearmanr(g,
                                                               p).correlation)

    jobs = list(product(range(n_genes), range(n_proteins)))

    # ====== multiprocessing ====== #
    mpi = MPI(jobs, func=_corr, ncpu=3, batch=len(jobs) // 3)
    all_correlations = sorted([i for i in mpi],
                              key=lambda scores:
                              (scores[-2] + scores[-1]) / 2)[::-1]
    return all_correlations

  # ******************** Protein analysis ******************** #
  def plot_protein_predicted_series(self,
                                    fontsize=10,
                                    proteins=None,
                                    y_true_new=None,
                                    y_pred_new=None,
                                    labels_new=None):
    if not self.is_semi_supervised:
      return self
    if proteins is not None:
      proteins = [standardize_protein_name(i).lower() for i in proteins]

    y_pred, y_true = self.y_pred, self.y_true
    if not self.is_binary_classes:
      y_true = log_norm(y_true, axis=1).numpy()
    # ====== override provided values ====== #
    labels = self.protein_name if labels_new is None else labels_new
    if y_true_new is not None:
      y_true = y_true_new
    if y_pred_new is not None:
      y_pred = y_pred_new

    n_protein = len(labels if proteins is None else
                    [i for i in labels if i.lower() in proteins])
    if n_protein == 0:
      return self
    colors = sns.color_palette(n_colors=2)

    #######
    if self.is_binary_classes:
      fig = plt.figure(figsize=(12, 12))
      from sklearn.metrics import confusion_matrix
      y_true = np.argmax(y_true, axis=-1)
      y_pred = np.argmax(y_pred, axis=-1)
      plot_confusion_matrix(cm=confusion_matrix(y_true, y_pred),
                            labels=labels,
                            colorbar=True,
                            fontsize=fontsize)
    #######
    else:
      fig = plt.figure(figsize=(12, n_protein * 4))
      cidx = 0
      for (name, pred, true) in zip(labels, y_pred.T, y_true.T):
        if proteins is not None and name.lower() not in proteins:
          continue
        cidx += 1

        assert pred.shape == true.shape
        ids = np.argsort(true)
        # TODO: handle different type of normalization for protein levels
        # e.g. if self.ynorm == 'prob':
        ax = plt.subplot(n_protein, 1, cidx)

        ax.plot(true[ids],
                linewidth=2,
                color=colors[0],
                label="[True]%s" % name)
        ax.plot(true[ids][0],
                linestyle='--',
                alpha=0.88,
                linewidth=1.2,
                color=colors[1],
                label="[Pred]%s" % name)
        ax.set_ylabel("Log-normalized true protein level", color=colors[0])
        ax.set_xlabel("Cell in sorted order of protein level")
        ax.tick_params(axis='y', colors=colors[0], labelcolor=colors[0])
        ax.set_title(name, fontsize=fontsize)
        ax.legend()

        ax = ax.twinx()
        ax.plot(pred[ids],
                linestyle='--',
                alpha=0.88,
                linewidth=1.2,
                color=colors[1],
                label="[Pred]%s" % name)
        ax.set_ylabel("Predicted protein response", color=colors[1])
        ax.tick_params(axis='y', colors=colors[1], labelcolor=colors[1])

    plt.suptitle(self.name, fontsize=fontsize + 4)
    with catch_warnings_ignore(UserWarning):
      plt.tight_layout(rect=[0, 0.03, 1, 0.96])
    self.add_figure('protein_series', fig)
    return self

  def plot_protein_scatter(self,
                           pca=False,
                           protein_name='CD4',
                           fig=None,
                           y_true_new=None,
                           y_pred_new=None,
                           labels_new=None):
    """ Plot comparison of protein in latent space, imputed and original
    input space """
    if not self.is_semi_supervised:
      return self

    fn_dim = fast_pca if pca else fast_tsne
    protein_name = standardize_protein_name(protein_name).strip().lower()
    fig_name = 'protein_scatter_%s' % protein_name

    labels = self.protein_name if labels_new is None else labels_new
    labels = [i.strip().lower() for i in labels]

    y_true = self.y_true if y_true_new is None else y_true_new
    y_pred = self.y_pred if y_pred_new is None else y_pred_new

    if protein_name not in labels:
      return self

    idx = [i for i, j in enumerate(labels) if protein_name in j][0]
    y_true = y_true[:, idx]
    y_pred = y_pred[:, idx]
    X, Z, V = self.X, self.Z, self.V

    if fig is None:
      fig = plot_figure(nrow=13, ncol=10)
    assert isinstance(fig, plt.Figure), \
    "fig must be instance of matplotlib.Figure"

    x = fn_dim(Z)
    ax = plot_scatter_heatmap(x, val=y_true, ax=321, grid=False, colorbar=True)
    ax.set_xlabel('Latent / ProteinOriginal')
    ax = plot_scatter_heatmap(x, val=y_pred, ax=322, grid=False, colorbar=True)
    ax.set_xlabel('Latent / ProteinPredicted')

    x = fn_dim(V)
    ax = plot_scatter_heatmap(x, val=y_true, ax=323, grid=False, colorbar=True)
    ax.set_xlabel('InputImputed / ProteinOriginal')
    ax = plot_scatter_heatmap(x, val=y_pred, ax=324, grid=False, colorbar=True)
    ax.set_xlabel('InputImputed / ProteinPredicted')

    x = fn_dim(X)
    ax = plot_scatter_heatmap(x, val=y_true, ax=325, grid=False, colorbar=True)
    ax.set_xlabel('InputOriginal / ProteinOriginal')
    ax = plot_scatter_heatmap(x, val=y_pred, ax=326, grid=False, colorbar=True)
    ax.set_xlabel('InputOriginal / ProteinPredicted')

    fig.suptitle(protein_name, fontsize=16)
    with catch_warnings_ignore(UserWarning):
      plt.tight_layout(rect=[0, 0.03, 1, 0.96])
    self.add_figure(fig_name, fig)
    return self

  # ******************** just summary everything  ******************** #
  def full_report(self, path, x_train=None, y_train=None, override=True):
    """
    path : `str`
      path to a folder
    """
    if not os.path.exists(path):
      os.mkdir(path)
    elif not os.path.isdir(path):
      raise ValueError("path to %s must be a folder" % path)
    else:
      if self.verbose:
        print("Override analysis at path:", path)
      shutil.rmtree(path)
      os.mkdir(path)

    score_path = os.path.join(path, 'scores.txt')
    self.save_scores(score_path)
    if self.verbose:
      print("Saved scores at:", score_path)

    self.plot_learning_curves(metric='loss')
    if self.is_semi_supervised:
      self.plot_protein_scatter(protein_name='CD4')
      self.plot_protein_scatter(protein_name='CD8')
      self.plot_protein_scatter(protein_name='CD45RA')
      self.plot_protein_scatter(protein_name='CD45RO')
      self.plot_protein_predicted_series()

    self.plot_correlation_top_pairs().plot_correlation_bottom_pairs()
    self.plot_correlation_marker_pairs(
        imputed=True).plot_correlation_marker_pairs(imputed=False)

    self.plot_classifier_F1(x_train, y_train)
    self.plot_latents_binary_scatter()
    self.plot_latents_distance_heatmap()
    self.plot_latents_protein_pairs()

    self.save_figures(path,
                      dpi=120,
                      separate_files=True,
                      clear_figures=True,
                      verbose=self.verbose)
    return self
